#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Reader for FAU/Erlangen's HDF5-based formats introduced with the pyccapt library."""

# pylint: disable=fixme

from typing import Dict
import os
import h5py
import numpy as np
import pandas as pd

from ase.data import atomic_numbers
from ifes_apt_tc_data_modeling.nexus.nx_ion import NxIon
from ifes_apt_tc_data_modeling.nexus.nx_field import NxField
from ifes_apt_tc_data_modeling.utils.utils import isotope_to_hash, \
    nuclide_hash_to_nuclide_list, MAX_NUMBER_OF_ATOMS_PER_ION
from ifes_apt_tc_data_modeling.utils.molecular_ions import get_chemical_symbols
# from ifes_apt_tc_data_modeling.utils.combinatorics import apply_combinatorics

# this implementation focuses on the following state of the pyccapt repository
# https://github.com/mmonajem/pyccapt/commit/e955beb4f2627befb8b4d26f2e74e4c52e00394e

# during the course of an atom probe measurement and analysis with FAU/Erlangen's Oxcart instrument
# several HDF5 files are generated with essentially two software tools. One is pyccapt which has a
# a control module plus a calibration module (whereby the voltage/bowl calibration and reconstruction
# is performed), the other one is a module/functionality to document ranging i.e. made ion type
# definitions. These results can be processed further with the FAU/Erlangen Matlab Atom Probe Toolbox;
# instructed as a set of Matlab live scripts this toolbox offers data analysis functionalities.
# Results obtained with this Atom Probe Toolbox are stored via an HDF5 file but currently not parsed
# by this pyccapt parser

# specific comments
# pyccapt/control
# an HDF5 file keeping relevant quantities

# pyccapt/calibration
# unfortunately the generated HDF5 file has internally no provenance information
# with which pyccapt version it was generated. Therefore, developers of pyccapt should rather
# write the content of the HDF5 file explicitly dset by dset e.g. using h5py instead
# of the pandas HDF5 dump convenience functionality
# of course pandas stores its own version but that is not conclusive enough to infer with
# which pyccapt version and most importantly from which other context the file was generated
# this is an aspect of the FAIR RDM principles which the pyccapt approach currently ignores
SUPPORTED_PYCCAPT_VERSION = "e955beb4f2627befb8b4d26f2e74e4c52e00394e"


def get_nuclide_hash_from_fau_list(elements, complexs, isotopes) -> np.ndarray:
    """Compute nuclide_hash from specific representation used at FAU/Erlangen."""
    # TODO:: add raise ValueError checks
    ivec = np.zeros((MAX_NUMBER_OF_ATOMS_PER_ION,), np.uint16)
    hashvector: list = []
    for idxj in np.arange(0, len(elements)):
        symbol = elements[idxj]
        if symbol in get_chemical_symbols():
            proton_number = atomic_numbers[symbol]
            neutron_number = isotopes[idxj] - proton_number
            hashvector.extend([isotope_to_hash(proton_number, neutron_number)] * complexs[idxj])
    ivec[0:len(hashvector)] = np.sort(np.asarray(hashvector, np.uint16), kind="stable")[::-1]
    return ivec


class ReadPyccaptControlFileFormat():
    """Read FAU/Erlangen pyccapt (controle module) HDF5 file format."""

    def __init__(self, file_path: str):
        if (file_path.lower().endswith(".h5") is False) and (file_path.lower().endswith(".hdf5") is False):
            raise ImportError("WARNING::HDF5 file incorrect file_path ending or file type!")
        self.file_path = file_path
        self.file_size = os.path.getsize(self.file_path)
        self.number_of_events = None
        self.version = SUPPORTED_PYCCAPT_VERSION

        # check that the formatting matches that of an pyccapt control module output HDF5 file
        with h5py.File(self.file_path, "r") as h5r:
            self.supported = 0  # voting-based
            required_groups = ["apt", "dld", "tdc"]
            for req_grpnm in required_groups:
                if req_grpnm in h5r.keys():
                    self.supported += 1
            if self.supported == 3:
                print(f"{self.file_path} is a supported pyccapt/control HDF5 file!")
            else:
                print(f"{self.file_path} is not a supported pyccapt/control HDF5 file!")
                return
        # parse out relevant pieces of information


class ReadPyccaptCalibrationFileFormat():
    """Read FAU/Erlangen pyccapt (calibration module) HDF5 file format."""

    def __init__(self, file_path: str):
        if (file_path.lower().endswith(".h5") is False) and (file_path.lower().endswith(".hdf5") is False):
            raise ImportError("WARNING::HDF5 file incorrect file_path ending or file type!")
        self.file_path = file_path
        self.file_size = os.path.getsize(self.file_path)
        self.number_of_events = None
        self.version = SUPPORTED_PYCCAPT_VERSION
        self.df: dict = {}

        with h5py.File(self.file_path, "r") as h5r:
            self.supported = 0  # voting-based
            required_entries = ["df",
                                "df/axis0", "df/axis1",
                                "df/block0_items", "df/block0_values",
                                "df/block1_items", "df/block1_values"]
            for entry in required_entries:
                if entry in h5r.keys():
                    self.supported += 1
            if self.supported == 7:
                print(f"{self.file_path} is a supported pyccapt/calibration HDF5 file!")
            else:
                print(f"{self.file_path} is not a supported pyccapt/calibration HDF5 file!")
                return

        self.df = pd.read_hdf(self.file_path)
        self.number_of_events = len(self.df)

    def get_named_quantities(self, term: str):
        """Get named quantities from dataframe."""
        if term in self.df.keys():
            return self.df[term]
        return None

    def get_reconstructed_positions(self):
        """Read xyz columns."""
        xyz = NxField()
        xyz.values = np.zeros([self.number_of_events, 3], np.float32)
        xyz.unit = "nm"

        dim = 0
        for quant in ["x (nm)", "y (nm)", "z (nm)"]:
            xyz.values[:, dim] = np.asarray(self.get_named_quantities(quant), np.float32)
            dim += 1
        return xyz

    def get_mass_to_charge_state_ratio(self):
        """Read (calibrated) mass-to-charge-state-ratio column."""

        m_n = NxField()
        m_n.values = np.zeros([self.number_of_events, 1], np.float32)
        m_n.unit = "Da"

        m_n.values[:, 0] = np.asarray(self.get_named_quantities("mc_c (Da)"), np.float32)
        return m_n


class ReadPyccaptRangingFileFormat():
    """Read FAU/Erlangen pyccapt (ranging module) HDF5 file format."""

    def __init__(self, file_path: str):
        if (file_path.lower().endswith(".h5") is False) and (file_path.lower().endswith(".hdf5") is False):
            raise ImportError("WARNING::HDF5 file incorrect file_path ending or file type!")
        self.file_path = file_path
        self.file_size = os.path.getsize(self.file_path)
        self.number_of_events = None
        self.version = SUPPORTED_PYCCAPT_VERSION
        self.df = None

        with h5py.File(self.file_path, "r") as h5r:
            self.supported = 0  # voting-based
            required_entries = ["df",
                                "df/axis0", "df/axis1",
                                "df/block0_items", "df/block0_values",
                                "df/block1_items", "df/block1_values",
                                "df/block2_items", "df/block2_values"]
            for entry in required_entries:
                if entry in h5r.keys():
                    self.supported += 1
            if self.supported == 9:
                print(f"{self.file_path} is a supported pyccapt/ranging HDF5 file!")
            else:
                print(f"{self.file_path} is not a supported pyccapt/ranging HDF5 file!")
                return

        self.df = pd.read_hdf(self.file_path)
        self.rng: Dict = {}
        self.rng["molecular_ions"] = []
        print(np.shape(self.df)[0])
        for idx in np.arange(0, np.shape(self.df)[0]):
            if isinstance(self.df.iloc[idx, 6], str) is True:
                if self.df.iloc[idx, 6] == "unranged":
                    continue

            ivec = get_nuclide_hash_from_fau_list(elements=self.df.iloc[idx, 6],
                                                  complexs=self.df.iloc[idx, 7],
                                                  isotopes=self.df.iloc[idx, 8])
            m_ion = NxIon()
            m_ion.nuclide_hash.values = ivec
            m_ion.nuclide_list.values = nuclide_hash_to_nuclide_list(ivec)
            m_ion.charge_state.values = np.int8(self.df.iloc[idx, 9])
            m_ion.add_range(self.df.iloc[idx, 3], self.df.iloc[idx, 4])
            m_ion.apply_combinatorics()
            # m_ion.report()
            self.rng["molecular_ions"].append(m_ion)
        print(f"{self.file_path} parsed successfully")



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Reader for *.analysis JSON-serialized objects from IVAS/Imago."""

# This is an example how to extract pieces of information from *.analysis files
# of Imago's IVAS analysis. Imago is the forerunner company of AMETEK/Cameca
# The example below shows how to extract ranging definitions.

import re
import xmltodict
import flatdict as fd
import numpy as np

from ase.data import atomic_numbers, chemical_symbols
from ifes_apt_tc_data_modeling.nexus.nx_ion import NxField, NxIon
from ifes_apt_tc_data_modeling.utils.definitions import \
    MAX_NUMBER_OF_ATOMS_PER_ION
from ifes_apt_tc_data_modeling.utils.molecular_ions import \
    get_chemical_symbols, isotope_to_hash


class ReadImagoAnalysisFileFormat():
    """Read *.analysis file (format), extract ranging definitions as an example."""

    def __init__(self, file_path: str):
        if (len(file_path) <= 9) or (file_path.lower().endswith(".analysis") is False):
            raise ImportError("WARNING::ANALYSIS file incorrect file_path ending or file type!")
        self.file_path = file_path
        self.imago: dict = {"ranges": {},
                            "ions": {},
                            "molecular_ions": []}
        self.read_imago_analysis_ranging_definitions()

    def read_imago_analysis_ranging_definitions(self):
        """Read *.analysis range file content."""
        # https://wilfried-grupe.de/Java_XMLEncoder.html
        # IVAS, the integrated visualization and analysis software/suite for
        # atom probe was a forerunner of the product that is now called APSuite
        # IVAS is a java application that uses JSON/XML object serialization
        # to store state files, many atom probe groups have still the resultant
        # state files within their projects, this function shows an example
        # how one can extract information from these files

        # unfortunately the serialized XML is in a format that when represented
        # in python mixes recurrent nested lists within dictionaries. Such a data
        # structure is not directly flattenable and hence many checks are required
        # I expect that this parser does not work out of the box for many
        # examples given that the formatting of object serialized content
        # depends heavily on the implementation of the host application (IVAS)
        # the main idea behind the example is to show that information can
        # be extracted and to motivate that nowadays one should use data
        # structures that are more conveniently parsable
        with open(self.file_path, "r", encoding="utf-8") as xmlf:
            xml = xmltodict.parse(xmlf.read())
            flt = fd.FlatDict(xml, "/")
            for entry in flt["java/object/void"]:
                # strategy is, walk the data structure and try to discard non-ranging content as early as possible
                for key, val in fd.FlatDict(entry, "/").items():
                    if (not isinstance(val, list)) or (key != "object/void"):
                        continue
                    for member in val:
                        if (not isinstance(member, dict)) \
                                or ("@method" not in member.keys()) \
                                or (member["@method"] != "add"):
                            continue
                        # print(">>>>>>>>>>>At the level of a molecular ion that can be so simple that it is just an element ion")
                        cand_dct = fd.FlatDict(member, "/")
                        # print(f">>>>> {cand_dct}")
                        all_reqs_exist = True
                        reqs = ["@method", "object/@id", "object/@class", "object/string", "object/boolean", "object/void"]
                        for req in reqs:
                            if req not in cand_dct.keys():
                                all_reqs_exist = False
                        if all_reqs_exist == False:
                            continue

                        if (not cand_dct["object/@id"].startswith("AtomDataRealRange")) \
                                or (cand_dct["object/@class"] != "com.imago.core.atomdata.AtomDataRealRange") \
                                or (not isinstance(cand_dct["object/void"], list)):
                            continue
                        for lst in cand_dct["object/void"]:
                            rng = fd.FlatDict(lst, "/")
                            element_symbol = []
                            mq = []
                            if "object/void/string" in rng.keys():
                                if rng["object/void/string"] in get_chemical_symbols():
                                    if "object/double" in rng.keys():
                                        mq = rng["object/double"][0:2]
                                        element_symbol.append(rng["object/void/string"])  # assuming multiplicity is one !
                            else:
                                if "object/void" in rng.keys():
                                    if isinstance(rng["object/void"], list):
                                        mq = rng["object/double"][0:2]
                                        element_symbol = []
                                        for block in rng["object/void"]:
                                            if isinstance(block, dict):
                                                if "@method" in block.keys() and "string" in block.keys() and "double" in block.keys():
                                                    if block["string"] in chemical_symbols:
                                                        for mult in np.arange(0, int(block["double"].split('.')[0])):
                                                            element_symbol.append(block["string"])
                            if (len(element_symbol) >= 1) and (len(mq) == 2):
                                # print(f"------------>{element_symbol}, {mq}")
                                ivec = []
                                for isotope in element_symbol:
                                    if isotope != "":
                                        prefix = re.findall("^[0-9]+", isotope)
                                        mass_number = 0
                                        if len(prefix) == 1:
                                            if int(prefix[0]) > 0:
                                                mass_number = int(prefix[0])
                                        suffix = re.findall("[0-9]+$", isotope)
                                        multiplier = 1
                                        if len(suffix) == 1:
                                            multiplier = int(suffix[0])
                                        symbol = isotope.replace(
                                            f"{mass_number}", "").replace(f"{multiplier}", "").replace(" ", "")
                                        if symbol in get_chemical_symbols():
                                            proton_number = atomic_numbers[symbol]
                                            neutron_number = 0
                                            if mass_number != 0:
                                                neutron_number = mass_number - proton_number
                                            ivec.extend([isotope_to_hash(proton_number, neutron_number)] * multiplier)
                                ivec = np.sort(np.asarray(ivec, np.uint16))[::-1]
                                ivector = np.zeros((MAX_NUMBER_OF_ATOMS_PER_ION,), np.uint16)
                                ivector[0:len(ivec)] = ivec

                                m_ion = NxIon(nuclide_hash=ivector, charge_state=0)
                                m_ion.add_range(float(mq[0]), float(mq[1]))
                                m_ion.comment = NxField(" ".join(element_symbol), "")
                                m_ion.apply_combinatorics()
                                m_ion.report()

                                self.imago["molecular_ions"].append(m_ion)
        print(f"{self.file_path} parsed successfully")



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""RRNG range file reader used by atom probe microscopists."""

# pylint: disable=too-many-branches,too-many-statements,duplicate-code

import re
import numpy as np

from ifes_apt_tc_data_modeling.nexus.nx_ion import NxField, NxIon, \
    try_to_reduce_to_unique_definitions
from ifes_apt_tc_data_modeling.utils.utils import \
    create_nuclide_hash, is_range_significant
from ifes_apt_tc_data_modeling.utils.definitions import MQ_EPSILON
from ifes_apt_tc_data_modeling.utils.molecular_ions import \
    get_chemical_symbols


def evaluate_rrng_range_line(i: int, line: str) -> dict:
    """Evaluate information content of a single range line."""
    # example line "Range7 = 42.8160 43.3110 vol:0.04543
    #     Al:1 O:1 Name: AlO Color:00FFFF"
    # according to DOI: 10.1007/978-1-4614-8721-0
    # mqmin, mqmax, vol, ion composition is required,
    #     name and color fields are optional
    info: dict = {"identifier": f"Range{i}",
                  "range": np.asarray([0., MQ_EPSILON], np.float64),
                  "atoms": [],
                  "volume": np.float64(0.),
                  "color": "",
                  "name": ""}

    tmp = re.split(r"[\s=]+", line)
    if len(tmp) < 6:
        # raise ValueError(f"Line {line} does not contain all required fields {len(tmp)}!")
        return None
    if is_range_significant(np.float64(tmp[1]), np.float64(tmp[2])) is False:
        # raise ValueError(f"Line {line} insignificant range!")
        return None
    info["range"] = np.asarray([tmp[1], tmp[2]], np.float64)

    if tmp[3].lower().startswith("vol:"):
        info["volume"] = np.float64(re.split(r":", tmp[3])[1])
    if (tmp[-1].lower().startswith("color:")) and \
       (len(re.split(r":", tmp[-1])[1]) == 6):
        info["color"] = "#" + re.split(r":", tmp[-1])[1]
    # HEX_COLOR_REGEX = r"^([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$"
    # replace r"^#( ...
    # regexp = re.compile(HEX_COLOR_REGEX)
    # if regexp.search(tmp[-1].split(r":")):

    for information in tmp[4:-1]:
        element_multiplicity = re.split(r":+", information)
        if len(element_multiplicity) != 2:
            raise ValueError(f"Line {line}, element multiplicity is not "
                             f"correctly formatted {len(element_multiplicity)}!")
        # skip vol, name, and color information
        if element_multiplicity[0].lower() == "name":
            info["name"] = f"{element_multiplicity[1]}"
            # this captures properly formatted ranges with keyword
            # name whose name value is then however not a chemical
            # symbol but some user-defined string
            # this is a common habit to define custom names
        elif element_multiplicity[0].lower() not in ["vol", "color"]:
            # pick up what is an element name
            symbol = element_multiplicity[0]
            if symbol not in get_chemical_symbols():
                # raise ValueError(f"WARNING::Line {line} contains an invalid chemical symbol {symbol}!")
                return info
            # if np.uint32(element_multiplicity[1]) <= 0:
                # raise ValueError(f"Line {line} zero or negative multiplicity !")
            if np.uint32(element_multiplicity[1]) >= 256:
                # raise ValueError(f"Line {line} unsupported high multiplicity "
                #                  f"{np.uint32(element_multiplicity)}!")
                return info
            info["atoms"] = np.append(info["atoms"],
                                      [symbol] * int(element_multiplicity[1]))
    return info


class ReadRrngFileFormat():
    """Read *.rrng file format."""

    def __init__(self, file_path: str, unique=False, verbose=False):
        """Initialize the class."""
        if (len(file_path) <= 5) or (file_path.lower().endswith(".rrng") is False):
            raise ImportError("WARNING::RRNG file incorrect file_path ending or file type!")
        self.file_path = file_path
        self.rrng: dict = {"ionnames": [],
                           "ranges": {},
                           "ions": {},
                           "molecular_ions": []}
        self.unique = unique
        self.verbose = verbose
        self.read_rrng()

    def read_rrng(self):
        """Read content of an RRNG range file."""
        with open(self.file_path, mode="r", encoding="utf8") as rrngf:
            txt = rrngf.read()

        txt = txt.replace("\r\n", "\n")  # windows to unix EOL conversion
        txt = txt.replace(",", ".")  # use decimal dots instead of comma
        txt_stripped = [line for line in txt.split("\n")
                        if line.strip() != "" and line.startswith("#") is False]
        del txt

        # see DOI: 10.1007/978-1-4899-7430-3 for further details to this
        # AMETEK/Cameca"s *.rrng file format

        # first, parse [Ions] section, which holds a list of element names
        # there are documented cases where experimentalists add custom strings
        # to specify ranges they consider special
        # these are loaded as user types
        # with nuclide_hash np.iinfo(np.uint16).max
        where = [idx for idx, element in
                 enumerate(txt_stripped) if element == "[Ions]"]
        if isinstance(where, list) is False:
            raise ValueError("Section [Ions] not found!")
        if len(where) != 1:
            raise ValueError("Section [Ions] not found or ambiguous!")
        current_line_id = where[0] + 1

        tmp = re.split(r"[\s=]+", txt_stripped[current_line_id])
        if len(tmp) != 2:
            raise ValueError(f"Line {txt_stripped[current_line_id]} [Ions]/Number line corrupted!")
        if tmp[0] != "Number":
            raise ValueError(f"Line {txt_stripped[current_line_id]} [Ions]/Number incorrectly formatted!")
        if tmp[1].isnumeric() is False:
            raise ValueError(f"Line {txt_stripped[current_line_id]} [Ions]/Number not a number!")
        number_of_ion_names = int(tmp[1])
        if number_of_ion_names <= 0:
            raise ValueError(f"Line {txt_stripped[current_line_id]} no ion names defined!")
        current_line_id += 1
        for i in np.arange(0, number_of_ion_names):
            tmp = re.split(r"[\s=]+", txt_stripped[current_line_id + i])
            if len(tmp) != 2:
                raise ValueError(f"Line {txt_stripped[current_line_id + i]} [Ions]/Ion line corrupted!")
            if tmp[0] != f"Ion{i + 1}":
                raise ValueError(f"Line {txt_stripped[current_line_id + i]} [Ions]/Ion incorrectly formatted!")
            if isinstance(tmp[1], str) is False:
                raise ValueError(f"Line {txt_stripped[current_line_id + i]} [Ions]/Name not a string!")
            self.rrng["ionnames"].append(tmp[1])

        # second, parse [Ranges] section
        where = [idx for idx, element in
                 enumerate(txt_stripped) if element == "[Ranges]"]
        if isinstance(where, list) is False:
            raise ValueError("Section [Ranges] not found!")
        if len(where) != 1:
            raise ValueError("Section [Ranges] not found or ambiguous!")
        current_line_id = where[0] + 1

        tmp = re.split(r"[\s=]+", txt_stripped[current_line_id])
        if len(tmp) != 2:
            raise ValueError(f"Line {txt_stripped[current_line_id]} [Ranges]/Number line corrupted!")
        if tmp[0] != "Number":
            raise ValueError(f"Line {txt_stripped[current_line_id]} [Ranges]/Number incorrectly formatted!")
        if tmp[1].isnumeric() is False:
            raise ValueError(f"Line {txt_stripped[current_line_id]} [Ranges]/Number not a number!")
        number_of_ranges = int(tmp[1])
        if number_of_ranges <= 0:
            raise ValueError(f"Line {txt_stripped[current_line_id]}  No ranges defined!")
        current_line_id += 1

        m_ions = []
        for jdx in np.arange(0, number_of_ranges):
            if self.verbose:
                print(f"{txt_stripped[current_line_id + jdx]}")
            dct = evaluate_rrng_range_line(jdx + 1, txt_stripped[current_line_id + jdx])
            if dct is None:
                print(f"WARNING::RRNG line {txt_stripped[current_line_id + jdx]} is corrupted!")
                continue

            m_ion = NxIon(nuclide_hash=create_nuclide_hash(dct["atoms"]), charge_state=0)
            m_ion.add_range(dct["range"][0], dct["range"][1])
            m_ion.comment = NxField(dct["name"], "")
            m_ions.append(m_ion)
            # this set may contain duplicates or overlapping ranges if ranging definitions are ambiguous like here https://doi.org/10.5281/zenodo.7788883

        if self.unique:
            unique_m_ions = try_to_reduce_to_unique_definitions(m_ions)
            print(f"Found {len(m_ions)} ranging definitions, performed reduction to {len(unique_m_ions)} unique ones")
        else:
            unique_m_ions = m_ions.copy()
            print(f"Found {len(m_ions)} ranging definitions, no reduction, {len(unique_m_ions)} remain.")
        del m_ions

        for m_ion in unique_m_ions:
            m_ion.apply_combinatorics()
            # m_ion.report()
            self.rrng["molecular_ions"].append(m_ion)
        print(f"{self.file_path} parsed successfully")



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""CSV file format reader as sometimes used for reporting POS file content."""

# pylint: disable=duplicate-code

import os
import numpy as np
import pandas as pd

from ifes_apt_tc_data_modeling.nexus.nx_field import NxField


class ReadCsvFileFormat():
    """Read CSV file assuming (n_ions, 4) like in POS."""

    def __init__(self, file_path: str):
        if (len(file_path) <= 4) or (file_path.lower().endswith(".csv") is False):
            raise ImportError("WARNING::CSV file incorrect file_path ending or file type!")
        self.file_path = file_path

        self.file_size = os.path.getsize(self.file_path)
        self.number_of_events = None
        self.version = None

        shp = np.shape(pd.read_csv(self.file_path))
        if shp[0] > 0 and shp[1] == 4:
            self.number_of_events = shp[0]
        else:
            raise ImportError("CSV file unsupported version because not formatted like POS!")

    def get_reconstructed_positions(self):
        """Read xyz columns."""

        xyz = NxField()
        xyz.values = np.zeros([self.number_of_events, 3], np.float32)
        xyz.unit = "nm"
        # there are too many assumption made here as to the content
        # in the csv file sure one could pass some configuration hints but
        # frankly there are nowadays much! better strategies to report
        # atom probe data than CSV, NeXus is one such, also csv files have
        # no magic number, de facto this works only because users know what
        # to expect in advance but how should a machine know this?
        for dim in [0, 1, 2]:
            xyz.values[:, dim] = pd.read_csv(self.file_path).iloc[:, dim]
        return xyz

    def get_mass_to_charge_state_ratio(self):
        """Read mass-to-charge-state-ratio column."""

        m_n = NxField()
        m_n.values = np.zeros([self.number_of_events, 1], np.float32)
        m_n.unit = "Da"
        # again such a strong assumption!
        # why reported in Da?
        # why in the third column
        # why at all a mass-to-charge-state-ratio value array?
        m_n.values[:, 0] = pd.read_csv(self.file_path).iloc[:, 3]
        return m_n



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Utility tool, analyze and combinatorics for (molecular) ions."""

# pylint: disable=too-many-arguments,too-many-instance-attributes,too-many-locals,too-many-nested-blocks,too-many-branches

import numpy as np
import radioactivedecay as rd

from ase.data import atomic_numbers, chemical_symbols
from ifes_apt_tc_data_modeling.utils.utils import \
    isotope_to_hash, hash_to_isotope, nuclide_hash_to_dict_keyword
from ifes_apt_tc_data_modeling.utils.nist_isotope_data import isotopes
from ifes_apt_tc_data_modeling.utils.definitions import \
    PRACTICAL_ABUNDANCE, PRACTICAL_ABUNDANCE_PRODUCT, \
    PRACTICAL_MIN_HALF_LIFE, SACRIFICE_ISOTOPIC_UNIQUENESS
VERBOSE = False


def get_chemical_symbols():
    """"Report only valid chemical symbols"""
    return chemical_symbols[1:]


class MolecularIonCandidate:
    """Define (molecular) ion build from nuclides."""

    def __init__(self,
                 ivec,
                 charge_state=0,
                 mass_sum=0.,
                 nat_abun_prod=0.,
                 min_half_life=np.inf):
        self.nuclide_hash = np.asarray(ivec, np.uint16)
        self.charge_state = np.int8(charge_state)
        self.mass = np.float64(mass_sum)
        self.abundance_product = np.float64(nat_abun_prod)
        self.shortest_half_life = np.float64(min_half_life)

    def unique_keyword(self):
        """Generate unique keyword."""
        keyword = f"{nuclide_hash_to_dict_keyword(np.sort(np.asarray(self.nuclide_hash, np.uint16), kind='stable')[::-1])}__{self.charge_state}"
        return keyword


class MolecularIonBuilder:
    """Class for holding properties of constructed molecular ions."""

    def __init__(self,
                 min_abundance=PRACTICAL_ABUNDANCE,
                 min_abundance_product=PRACTICAL_ABUNDANCE_PRODUCT,
                 min_half_life=PRACTICAL_MIN_HALF_LIFE,
                 sacrifice_uniqueness=SACRIFICE_ISOTOPIC_UNIQUENESS,
                 verbose=VERBOSE):
        self.nuclides = np.asarray([], np.uint16)
        self.element_isotopes = {}
        self.nuclide_mass = {}
        self.nuclide_abundance = {}
        self.nuclide_stable = {}  # observationally stable
        self.nuclide_unclear = {}  # unclear halflife
        self.nuclide_halflife = {}
        self.candidates = []
        self.parms = {"min_abundance": min_abundance,
                      "min_abundance_product": min_abundance_product,
                      "min_half_life": min_half_life,
                      "sacrifice_isotopic_uniqueness": sacrifice_uniqueness,
                      "verbose": verbose}

        for symbol, atomic_number in atomic_numbers.items():
            if symbol != "X":
                # assume that data from ase take preference
                # if half-life data are available in radioactive decay library
                # take these instead, if all fails mark an unclear_half_life == True
                element_isotopes = []
                for mass_number in isotopes[atomic_number]:
                    half_life = np.inf
                    observationally_stable = False
                    unclear_half_life = False

                    # test if half-life data available
                    trial_nuclide_name = f"{symbol}-{mass_number}"
                    try:
                        tmp = rd.Nuclide(trial_nuclide_name)
                    except ValueError:
                        tmp = None
                    if tmp is not None:
                        half_life = tmp.half_life()
                        if np.isinf(half_life):
                            observationally_stable = True
                            # these ions are always taken as they
                            # are most relevant for practical
                            # atom probe experiments
                        else:
                            if half_life < self.parms["min_half_life"]:
                                # ignore practically short living ions
                                continue
                    else:
                        continue
                        # do not consider exotic isotopes with unclear
                        # half-life as they are likely anyway irrelevant
                        # for practical atom probe experiments
                        # half_life = np.nan
                        # unclear_half_life = True

                    # get ase abundance data
                    n_protons = atomic_numbers[symbol]
                    n_neutrons = mass_number - n_protons
                    mass = isotopes[atomic_numbers[symbol]][mass_number]["mass"]
                    abundance = isotopes[atomic_numbers[symbol]][mass_number]["composition"]
                    hashvalue = isotope_to_hash(int(n_protons), int(n_neutrons))
                    if hashvalue != 0:
                        self.nuclides = np.append(self.nuclides, hashvalue)
                        self.nuclide_mass[hashvalue] = np.float64(mass)
                        self.nuclide_abundance[hashvalue] = np.float64(abundance)
                        self.nuclide_stable[hashvalue] = observationally_stable
                        self.nuclide_unclear[hashvalue] = unclear_half_life
                        self.nuclide_halflife[hashvalue] = half_life
                        element_isotopes = np.append(element_isotopes, hashvalue)
                self.element_isotopes[atomic_number] = np.sort(
                    np.asarray(element_isotopes, np.uint16), kind="stable")[::-1]
        self.nuclides = np.sort(self.nuclides, kind="stable")[::-1]
        if self.parms["verbose"] is True:
            print(f"MolecularIonBuilder initialized with {len(self.nuclides)} nuclides")

    def get_element_isotopes(self, hashvalue):
        """List of hashvalues all isotopes of element specified by hashvalue."""
        return self.element_isotopes[hash_to_isotope(hashvalue)[0]]

    def get_isotope_mass_sum(self, nuclide_arr):
        """Evaluate cumulated atomic_mass of isotopes in ivec."""
        # assuming no relativistic effects or other quantum effects
        # mass loss due to charge_state considered insignificant
        mass = 0.
        for hashvalue in nuclide_arr:
            if hashvalue != 0:
                mass += self.nuclide_mass[hashvalue]
        return mass

    def get_natural_abundance_product(self, nuclide_arr):
        """Get natural abundance product."""
        abun_prod = 1.
        for hashvalue in nuclide_arr:
            if hashvalue != 0:
                abun_prod *= self.nuclide_abundance[hashvalue]
        return abun_prod

    def get_shortest_half_life(self, nuclide_arr):
        """Get shortest half life for set of nuclides."""
        min_half_life = self.parms["min_half_life"]
        for hashvalue in nuclide_arr:
            if hashvalue != 0:
                if self.nuclide_halflife[hashvalue] != np.nan:
                    min_half_life = np.min((min_half_life, self.nuclide_halflife[hashvalue]))
                    # if the min_half_life is np.inf than we know that every nuclide in the
                    # molecular ion is observationally stable
                    # if not we know that considering this molecular ion might not be a good
                    # idea or only necessary in very few cases because even if we were
                    # to find such a combination of nuclides at least one would decay in observable
                    # time and this might possibly hint that studying this molecular is tricky
                else:
                    # if we do not information about the half-life it is very likely that
                    # this is an exotic nuclide likely never found in the wild
                    return np.nan
        return min_half_life

    def combinatorics(self, hash_arr, low, high):
        """Combinatorial analysis which (molecular) elements match within [low, high]."""
        # RNG/RRNG/ENV range files store (molecular) ion information for each range
        # BUT neither nuclide not charge state information, here we try to recover
        # both if possible or list all possible combinations
        # correspondingly this allows yield
        # only an nuclide_hash whose hashvalues have ALL in common
        # that the number of neutrons is 0, i.e. their hashvalue is the atomic_number
        # hash_arr is an nuclide_hash/ivec with such hashvalues
        # as an example the molecular ion C:2 H:1 will map to 6, 6, 1
        max_depth = 0  # number of non-zero entries in hash_arr
        for hashvalue in hash_arr:
            if hashvalue != 0:
                max_depth += 1
        if self.parms["verbose"] is True:
            print(f"Maximum recursion depth {max_depth}")
        self.candidates = []
        if self.parms["verbose"] is True:
            print(hash_arr)

        if max_depth > 0:
            depth = 0
            ith_nuclides = self.get_element_isotopes(hash_arr[depth])
            cand_arr_curr = []  # combinatorially add nuclides while recursing deeper
            self.iterate_molecular_ion(
                hash_arr, ith_nuclides, cand_arr_curr,
                depth, max_depth, low, high)
            if self.parms["verbose"] is True:
                print(f"Found {len(self.candidates)} candidates!")
                for obj in self.candidates:
                    print(f"{obj.nuclide_hash}, {obj.charge_state}, {obj.shortest_half_life}")
            return self.try_to_reduce_to_unique_solution()
            # will return a tuple of charge_state and list of relevant_candidates
        return (0, [])

    def iterate_molecular_ion(self,
                              hash_arr, jth_nuclides, cand_arr_prev,
                              i, max_n, low, high):
        """Recursive analysis of combinatorics on molecular ions."""
        if i < (max_n - 1):
            for nuclide in jth_nuclides:
                ixxth_nuclides = self.get_element_isotopes(hash_arr[i + 1])
                cand_arr_curr = np.append(cand_arr_prev, nuclide)
                self.iterate_molecular_ion(
                    hash_arr, ixxth_nuclides, cand_arr_curr,
                    i + 1, max_n, low, high)
        elif i == (max_n - 1):
            for nuclide in jth_nuclides:
                cand_arr_curr = np.append(cand_arr_prev, nuclide)
                # by this design the ivec does not necessarily remain ordered

                new_mass = self.get_isotope_mass_sum(cand_arr_curr)
                new_abun_prod = self.get_natural_abundance_product(cand_arr_curr)
                new_halflife = self.get_shortest_half_life(cand_arr_curr)

                for new_chrg in np.arange(1, 8):
                    mass_to_charge = new_mass / new_chrg
                    if mass_to_charge < low:
                        break
                    # we can break the entire charge state generation here already
                    # instead of continue because already the current mq is left out
                    # of interval [mqmin, mqmax]
                    # and all mq in the next iterations will result in even lower
                    # mass-to-charge you walk to the left increasing your
                    # distance to the left bound
                    if mass_to_charge > high:
                        # can be optimized and broken out of earlier
                        # if testing first chrg == 1
                        # and then chrg == APTMOLECULARION_MAX_CHANGE
                        continue
                        # must not be break here because with adding more charge
                        # we usually walk from right to left eventually into
                        # [low, high] !
                    # molecular ion is within user-specified bounds
                    self.candidates.append(
                        MolecularIonCandidate(cand_arr_curr,
                                              new_chrg,
                                              new_mass,
                                              new_abun_prod,
                                              new_halflife))
                # do not start a new recursion
        else:
            # break the recursion
            return

    def get_relevant(self):
        """Identify relevant candidates."""
        relevant = {}
        for cand in self.candidates:
            if cand.abundance_product >= self.parms["min_abundance_product"]:
                if not np.isnan(cand.shortest_half_life):
                    # don't dare to test np.isnan(cand.shortest_half_life) is False
                    if cand.shortest_half_life >= self.parms["min_half_life"]:
                        keyword = cand.unique_keyword()
                        if keyword not in relevant:
                            relevant[keyword] = cand

        if self.parms["verbose"] is True:
            print(f"Reduced set to {len(relevant.keys())} relevant candidates...")
            for key in relevant:
                print(key)
        relevant_candidates = []
        for key, obj in relevant.items():
            relevant_candidates.append(obj)
        return (relevant, relevant_candidates)

    def try_to_reduce_to_unique_solution(self):
        """Heuristics to identify if current candidates are unique."""
        if self.parms["verbose"] is True:
            print(f"Reduce set of {len(self.candidates)} candidates to a unique...")
        relevant, relevant_candidates = self.get_relevant()
        if len(relevant) == 0:
            if self.parms["verbose"] is True:
                print("WARNING::No relevant candidate meets all criteria!")
                print("WARNING::No solution possible for given criteria!")
            return (0, relevant_candidates)
        if len(relevant) == 1:
            if self.parms["verbose"] is True:
                print("One relevant candidate which meets all criteria")
            keywords = []
            for key in relevant:
                if isinstance(key, str):
                    keywords.append(key)
            assert len(keywords) >= 1, "List of relevant keywords is empty!"
            return (relevant[keywords[0]].charge_state, relevant_candidates)

        if self.parms["verbose"] is True:
            print("Multiple relevant candidates meet all selection criteria")
        keywords = []
        for key in relevant:
            if isinstance(key, str):
                keywords.append(key)
        assert len(keywords) >= 1, "List of relevant keywords is empty!"
        charge_state = relevant[keywords[0]].charge_state
        for key, val in relevant.items():
            if val.charge_state == charge_state:
                continue
            if self.parms["verbose"] is True:
                print("WARNING::Multiple relevant candidates differ in charge_state!")
                print("WARNING::No unique solution possible for given criteria!")
            return (0, relevant_candidates)
        # not returned yet, so all relevant candidates have the same charge_state
        if self.parms["verbose"] is True:
            print(f"Multiple relevant candidates have all the same charge_state {charge_state}")
        if self.parms["sacrifice_isotopic_uniqueness"] is True:
            return (charge_state, relevant_candidates)
        if self.parms["verbose"] is True:
            print("WARNING::Multiple relevant candidates differ in isotopes!")
            print("WARNING::But these have the same charge_state!")
            print("WARNING::No unique solution possible for given criteria!")
        return (0, relevant_candidates)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Generic definitions when working with molecular ions useful for atom probe."""

import numpy as np

# restrict the number distinguished ion types
MAX_NUMBER_OF_ION_SPECIES = 256
# restrict number of atoms for molecular ion fragments
MAX_NUMBER_OF_ATOMS_PER_ION = 32
# practical and required minimum mass-resolution Da or atomic mass unit (amu)
MQ_EPSILON = np.float64(1. / 2000.)


# do not consider isotopes with a very low natural abundance
PRACTICAL_ABUNDANCE = 1.0e-6  # 0.  # 1.0e-6
# do not consider candidate isotopically different molecular ions
# if their natural abundance product is too low
PRACTICAL_ABUNDANCE_PRODUCT = 0.  # 1.0e-6  # 0.  # 1.0e-12
# do consider too shortliving isotopes
PRACTICAL_MIN_HALF_LIFE = np.inf
# many examples of ranges are not constrainted strongly enough so that
# there are many isotopes (many of which admittedly hypothetical) ones
# which are within the range, this option lifts the constraint that
# there should be only one set of isotopically different molecular ions
# and if these have all these same charge it is assumed this is the
# charge of the molecular ion
# strictly speaking however one would have to rule out every possible
# molecular ion candidate (which is non-trivial and maybe not even
# with first principles theory possible...
SACRIFICE_ISOTOPIC_UNIQUENESS = True



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Utilities for working with molecular ions in atom probe microscopy."""

from typing import Tuple
import re
import numpy as np

from ase.data import atomic_numbers, chemical_symbols
from ifes_apt_tc_data_modeling.utils.nist_isotope_data import isotopes
from ifes_apt_tc_data_modeling.utils.definitions import \
    MAX_NUMBER_OF_ATOMS_PER_ION, MQ_EPSILON


def get_smart_chemical_symbols():
    """Organize element symbols such that search H does not match He."""
    priority_queue = []
    for symbol in chemical_symbols:
        if len(symbol) == 2:
            priority_queue.append(symbol)
    for symbol in chemical_symbols:
        if symbol != "X" and len(symbol) == 1:
            priority_queue.append(symbol)
    return priority_queue


def isotope_to_hash(proton_number: int = 0,
                    neutron_number: int = 0) -> int:
    """Encode an isotope to a hashvalue."""
    if (0 <= proton_number < 256) and (0 <= neutron_number < 256):
        return int(np.uint16(proton_number) + (np.uint16(256) * np.uint16(neutron_number)))
    return 0


def hash_to_isotope(hashvalue: int = 0) -> Tuple[int, int]:
    """Decode a hashvalue to an isotope."""
    # assert isinstance(hashvalue, int), \
    #     "Argument hashvalue needs to be integer!"
    if 0 <= hashvalue <= int(np.iinfo(np.uint16).max):
        neutron_number = np.uint16(np.uint16(hashvalue) / np.uint16(256))
        proton_number = np.uint16(np.uint16(hashvalue) - neutron_number * np.uint16(256))
        return (int(proton_number), int(neutron_number))
    return (0, 0)


def create_nuclide_hash(building_blocks: list) -> np.ndarray:
    """Create specifically-shaped array of isotope hashvalues."""
    # building_blocks are usually names of elements in the periodic table
    # if not we assume the ion is special such as user type or plain words
    # a typical expected test case is
    # create_nuclide_hash(["Fe", "Fe", "O", "O", "O"])
    ivec = np.zeros((MAX_NUMBER_OF_ATOMS_PER_ION,), np.uint16)
    if 0 < len(building_blocks) <= MAX_NUMBER_OF_ATOMS_PER_ION:
        symbol_to_proton_number = atomic_numbers
        hashvector = []
        for block in building_blocks:
            if isinstance(block, str) and block != "":
                if block.count("-") == 0:  # an element
                    if (block not in symbol_to_proton_number) or (block == "X"):
                        return ivec
                    hashvector.append(isotope_to_hash(symbol_to_proton_number[block], 0))
                elif block.count("-") == 1:
                    symb_mass = block.split("-")
                    if (len(symb_mass) != 2) or (symb_mass[0] not in symbol_to_proton_number) or (symb_mass[0] == "X"):
                        print(f"WARNING:: {block} is not properly formatted <symbol>-<mass_number>!")
                        return ivec
                    proton_number = symbol_to_proton_number[symb_mass[0]]
                    mass_number = int(symb_mass[1])
                    neutron_number = mass_number - proton_number
                    if (proton_number in isotopes) and (mass_number in isotopes[proton_number]):
                        hashvector.append(isotope_to_hash(proton_number, neutron_number))
        ivec[0:len(hashvector)] = np.sort(np.asarray(hashvector, np.uint16), kind="stable")[::-1]
    return ivec


def nuclide_hash_to_nuclide_list(ivec: np.ndarray) -> np.ndarray:
    """Create a NeXus NXion nuclide list."""
    nuclide_list = np.zeros((MAX_NUMBER_OF_ATOMS_PER_ION, 2), np.uint16)
    if np.shape(ivec) == (MAX_NUMBER_OF_ATOMS_PER_ION,):
        for idx in np.arange(0, MAX_NUMBER_OF_ATOMS_PER_ION):
            if ivec[idx] != 0:
                n_protons, n_neutrons = hash_to_isotope(int(ivec[idx]))
                if n_neutrons != 0:
                    nuclide_list[idx, 0] = n_protons + n_neutrons
                nuclide_list[idx, 1] = n_protons
        return nuclide_list
    print(f"WARNING:: Argument nuclide_hash needs to be shaped ({MAX_NUMBER_OF_ATOMS_PER_ION},) !")
    return nuclide_list


def nuclide_hash_to_dict_keyword(ivec: np.ndarray) -> str:
    """Create keyword for dictionary from nuclide_hash."""
    if len(ivec) <= MAX_NUMBER_OF_ATOMS_PER_ION:
        lst = []
        for hashvalue in ivec:
            if hashvalue != 0:
                lst.append(f"{hashvalue}")
        if len(lst) > 0:
            return "_".join(lst)
    return "0"  # "_".join(np.asarray(np.zeros((MAX_NUMBER_OF_ATOMS_PER_ION,)), np.uint16))


def nuclide_hash_to_human_readable_name(ivec: np.ndarray, charge_state: np.int8) -> str:
    """Get human-readable name from an nuclide_hash."""
    if len(ivec) <= MAX_NUMBER_OF_ATOMS_PER_ION:
        human_readable = ""
        for hashvalue in ivec:
            if hashvalue != 0:
                protons, neutrons = hash_to_isotope(int(hashvalue))
                if neutrons > 0:
                    human_readable += f"{protons + neutrons}{chemical_symbols[protons]}"
                else:
                    human_readable += f"{chemical_symbols[protons]}"
                human_readable += " "
        if 0 < charge_state < 8:
            human_readable += "+" * charge_state
        elif -8 < charge_state < 0:
            human_readable += "-" * -charge_state
        else:
            human_readable = human_readable.rstrip()
        return human_readable
    return "unknown_iontype"


def is_range_overlapping(interval: np.ndarray,
                         interval_set: np.ndarray) -> bool:
    """Check if interval overlaps within with members of interval set."""
    if (np.shape(interval) == (2,)) and (np.shape(interval_set)[0] >= 1) \
            and (np.shape(interval_set)[1] == 2):
        # interval = np.array([53.789, 54.343])
        # interval_set = np.array([[27.778, 28.33]])  # for testing purposes
        left_and_right_delta = np.zeros([np.shape(interval_set)[0], 2], bool)
        left_and_right_delta[:, 0] = (interval_set[:, 0] - interval[1]) \
            > MQ_EPSILON
        left_and_right_delta[:, 1] = (interval[0] - interval_set[:, 1]) \
            > MQ_EPSILON
        no_overlap = np.array([np.any(i) for i in left_and_right_delta])
        if np.all(no_overlap):
            return False
        return True
    return False


def is_range_significant(left: np.float64, right: np.float64) -> bool:
    """Check if inclusive interval bounds [left, right] span a finite range."""
    if (np.float64(0.) <= left) and (np.float64(0.) <= right):
        if (right - left) >= MQ_EPSILON:
            return True
    return False


def is_convertible_to_isotope_hash(symbol: str):
    """Check if human_readable symbol is convertible into nuclide hash tribool."""
    case = re.search("^([A-Z])([a-z])?(-)([0-9]+)$", symbol)
    if case is None:  # eventually element case e.g. "K"
        if not isinstance(symbol, str):
            raise ValueError("Argument symbol needs to be a string !")
        if symbol not in get_smart_chemical_symbols():
            raise ValueError(f"Symbol needs to be in {get_smart_chemical_symbols()}!")
        return 1
    # alternative case eventually specific nuclide e.g. "K-40"
    symb_mass = symbol.split("-")
    if len(symb_mass) != 2:
        raise TypeError("Argument symbol is not properly formatted <symbol>-<mass_number>!")
    if len(symb_mass[0]) != 1 and len(symb_mass[0]) != 2:
        raise ValueError("Argument symbol is not properly formatted <symbol>-<mass_number>!")
    if len(symb_mass[1]) <= 0:
        raise ValueError(f"Argument symbol {symb_mass[1]} needs to be a physical mass number!")
    if symb_mass[0] not in get_smart_chemical_symbols():
        raise ValueError(f"{symb_mass[0]} is not a symbol in {get_smart_chemical_symbols()}!")
    if int(symb_mass[1]) not in isotopes[atomic_numbers[symb_mass[0]]].keys():
        raise ValueError(f"No value for isotopes[atomic_numbers[{symb_mass[0]}][{int(symb_mass[1])}] exists!")
    return 2


def element_or_nuclide_to_hash(symbol: str):
    """Converts an element symbol (e.g. K) or nuclide (K-40) to nuclide hash."""
    # consider moving this to the ifes_apt_tc_data_modeling library
    case = is_convertible_to_isotope_hash(symbol)
    if case == 1:
        return isotope_to_hash(atomic_numbers[symbol], 0)
    if case == 2:
        symb_mass = symbol.split("-")
        return isotope_to_hash(atomic_numbers[symb_mass[0]],
                               int(symb_mass[1]) - atomic_numbers[symb_mass[0]])
    return isotope_to_hash(0, 0)


def symbol_lst_to_matrix_of_nuclide_vector(method: str = "resolve_all",
                                           symbol_lst: list = [],
                                           **kwargs):
    """Parse human-readable element/isotope names to paraprobe."""
    # method = "resolve_all"
    # method = "resolve_unknown"
    # method = "resolve_element"
    # method = "resolve_isotope"
    # method = "resolve_ion"
    # symbol_lst = [["K", "C"], ["U-238", "H-2"]]
    # symbol_lst = [["K-40", "C-14"], ["U-238", "H-2"]]  # one list per mol. ion
    # kwargs_charge_lst = [+1, +3]  # one charge per ion
    # resolve_all and resolve_unknown are the special cases
    # where we do not need to specify a matrix of isotope vectors
    if method in ("resolve_all", "resolve_unknown"):
        return (None, None, None)

    supported = ("resolve_element", "resolve_ion", "resolve_isotope")
    if method not in supported:
        raise ValueError(f"Argument method needs to be in {supported} !")
    if not isinstance(symbol_lst, list) and all(isinstance(lst, list) for lst in symbol_lst):
        raise TypeError("Argument symbol_lst must be a list of lists!")
    if not all(len(np.shape(lst)) == 1 for lst in symbol_lst):
        raise ValueError("One list in argument symbol_lst is not a 1d list or an empty list!")
    if not all(np.shape(lst)[0] >= 1 for lst in symbol_lst):
        raise ValueError("One list in argument symbol_lst is not a 1d list or an empty list!")
    matrix = np.zeros([len(symbol_lst), MAX_NUMBER_OF_ATOMS_PER_ION])
    charge = []
    if (method == "resolve_ion") and ("charge_lst" in kwargs):
        if not isinstance(kwargs["charge_lst"], list):
            raise TypeError("Keyword argument charge_lst must be a list of lists!")
        if not all(isinstance(val, int) for val in kwargs["charge_lst"]):
            raise ValueError("Keyword argument charge_lst needs to be a list of int !")
        if np.shape(symbol_lst)[0] != np.shape(kwargs["charge_lst"])[0]:
            raise ValueError("Argument symbol_lst and keyword argument charge_lst need to have the same length !")
    for idx, lst in enumerate(symbol_lst):
        ivec = np.zeros([1, MAX_NUMBER_OF_ATOMS_PER_ION])
        if lst == []:
            raise ValueError("Argument molecular ion must not be an empty list!")
        if len(lst) > MAX_NUMBER_OF_ATOMS_PER_ION:
            raise ValueError(f"Argument molecular ion must not contain more than "
                             f"{MAX_NUMBER_OF_ATOMS_PER_ION} entries!")
        jdx = 0
        for symbol in lst:
            if method == "resolve_element":
                if symbol in atomic_numbers:
                    ivec[0, jdx] = isotope_to_hash(atomic_numbers[symbol], 0)  # do not encode isotope information
                    jdx += 1
                else:
                    # it might be that we have a specific nuclide e.g. K-40 try to parse element symbol
                    candidate = symbol.split("-", 1)[0]
                    if candidate not in atomic_numbers:
                        raise KeyError(f"symbol_lst[{idx}] candidate does not specify an element!")
                    ivec[0, jdx] = isotope_to_hash(atomic_numbers[candidate], 0)
                    jdx += 1
            else:  # "resolve_isotope", "resolve_ion":
                ivec[0, jdx] = element_or_nuclide_to_hash(symbol)
                jdx += 1
        matrix[idx, :] = ivec[0, :]
        if method == "resolve_ion":
            charge.append(kwargs["charge_lst"][idx])
    if charge == []:
        return (method, np.asarray(matrix, np.uint16), None)
    return (method, np.asarray(matrix, np.uint16), np.asarray(charge, np.int8))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""RNG range file reader used by atom probe microscopists."""

# pylint: disable=duplicate-code

import re
import numpy as np

from ifes_apt_tc_data_modeling.nexus.nx_ion import NxField, NxIon
from ifes_apt_tc_data_modeling.utils.utils import \
    create_nuclide_hash, is_range_significant
from ifes_apt_tc_data_modeling.utils.definitions import MQ_EPSILON
from ifes_apt_tc_data_modeling.utils.molecular_ions import get_chemical_symbols


# there are specific examples for unusual range files here:
# https://hg.sr.ht/~mycae/libatomprobe/browse/test/samples/ranges?rev=tip


def evaluate_rng_range_line(
        i: int, line: str, column_id_to_label: dict, n_columns: int) -> dict:
    """Represent information content of a single range line."""
    # example line: ". 107.7240 108.0960 1 0 0 0 0 0 0 0 0 0 3 0 0 0"
    info: dict = {"identifier": f"Range{i}",
                  "range": np.asarray([0., MQ_EPSILON], np.float64),
                  "atoms": [],
                  "volume": np.float64(0.),
                  "color": "",
                  "name": ""}

    tmp = re.split(r"\s+", line)
    if len(tmp) != n_columns:
        raise ValueError(f"Line {line} inconsistent number columns {len(tmp)}!")
    if tmp[0] != ".":
        raise ValueError(f"Line {line} has inconsistent line prefix!")
    if is_range_significant(np.float64(tmp[1]), np.float64(tmp[2])) is False:
        # raise ValueError(f"Line {line} insignificant range!")
        return info
    info["range"] = np.asarray([tmp[1], tmp[2]], np.float64)

    # line encodes multiplicity of element via array of multiplicity counts
    element_multiplicity = np.asarray(tmp[3:len(tmp)], np.uint32)
    if np.sum(element_multiplicity) < 0:
        # raise ValueError(f"Line {line} no element counts!")
        return info
    if np.sum(element_multiplicity) > 0:
        for jdx in np.arange(0, len(element_multiplicity)):
            if element_multiplicity[jdx] < 0:
                # raise ValueError(f"Line {line} no negative element counts!")
                raise ValueError(f"element_multiplicity[jdx] {element_multiplicity[jdx]} needs to be positive!")
            if element_multiplicity[jdx] > 0:
                symbol = column_id_to_label[jdx + 1]
                if symbol in get_chemical_symbols():
                    info["atoms"] = np.append(info["atoms"],
                                              [column_id_to_label[jdx + 1]] * int(element_multiplicity[jdx]))
                else:
                    info["name"] = symbol
                    info["atoms"] = []  # will map to unknown type

    # color for RNG files can only be decoded by
    # loading the color of elements and polyatomic extensions
    # and then check to which category (element or polyatomic) a range
    # belongs and then take this color, there is almost no way
    # to make something so simple for disentangled
    return info


def evaluate_rng_ion_type_header(line: str) -> dict:
    """Represent information content in the key header line."""
    # line = "------------------- Fe Mg Al Mn Si V C Ga Ti Ca O Na Co H"
    # line = "---- a"
    # line = "----------------- Sc Fe O C Al Si Cr H unknown"
    info: dict = {"column_id_to_label": {}}
    tmp = re.split(r"\s+", line)
    if len(tmp) == 0:
        raise ValueError(f"Line {line} does not contain iontype labels {len(tmp)}!")
    for idx in np.arange(1, len(tmp)):
        info["column_id_to_label"][idx] = tmp[idx]
    return info


class ReadRngFileFormat():
    """Read *.rng file format."""

    def __init__(self, file_path: str):
        if (len(file_path) <= 4) or (file_path.lower().endswith(".rng") is False):
            raise ImportError("WARNING::RNG file incorrect file_path ending or file type!")
        self.file_path = file_path
        self.rng: dict = {"ranges": {},
                          "ions": {},
                          "molecular_ions": []}
        self.read_rng()

    def read_rng(self):
        """Read RNG range file content."""
        with open(self.file_path, mode="r", encoding="utf8") as rngf:
            txt = rngf.read()

        txt = txt.replace("\r\n", "\n")  # windows to unix EOL conversion
        txt = txt.replace(",", ".")  # use decimal dots instead of comma
        txt_stripped = [line for line in txt.split("\n")
                        if line.strip() != "" and line.startswith("#") is False]
        del txt

        # see DOI: 10.1007/978-1-4899-7430-3 for further details to this
        # Oak Ridge National Lab / Oxford *.rng file format
        # only the first ------ line is relevant
        # it details all ion labels aka ions
        # AMETEK"s IVAS/APSuite-specific trailing
        # polyatomic extension is redundant info

        tmp = None
        current_line_id = int(0)  # search key header line
        for line in txt_stripped:
            tmp = re.search(r"----", line)
            if tmp is None:
                current_line_id += int(1)
            else:
                break
        if tmp is None:
            raise ValueError("RNG file does not contain key header line!")

        header = evaluate_rng_ion_type_header(txt_stripped[current_line_id])

        tmp = re.split(r"\s+", txt_stripped[0])
        if tmp[0].isnumeric() is False:
            raise ValueError(f"Line {txt_stripped[0]} number of species corrupted!")
        n_element_symbols = int(tmp[0])
        if n_element_symbols < 0:
            raise ValueError(f"Line {txt_stripped[0]} no species defined!")
        if tmp[1].isnumeric() is False:
            raise ValueError(f"Line {txt_stripped[0]} number of ranges corrupted!")
        n_ranges = int(tmp[1])
        if n_ranges < 0:
            raise ValueError(f"Line {txt_stripped[0]} no ranges defined!")

        for idx in np.arange(current_line_id + 1, current_line_id + 1 + n_ranges):
            dct = evaluate_rng_range_line(
                idx - current_line_id, txt_stripped[idx],
                header["column_id_to_label"],
                n_element_symbols + 3)
            if dct is None:
                print(f"WARNING::RNG line {txt_stripped[idx]} is corrupted!")
                continue

            m_ion = NxIon(nuclide_hash=create_nuclide_hash(dct["atoms"]), charge_state=0)
            m_ion.add_range(dct["range"][0], dct["range"][1])
            m_ion.comment = NxField(dct["name"], "")
            m_ion.apply_combinatorics()
            # m_ion.report()

            self.rng["molecular_ions"].append(m_ion)
        print(f"{self.file_path} parsed successfully")



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""AMETEK APT(6) data exchange file reader used by atom probe microscopists."""

import numpy as np

APT_SECTION_NAME_MAX_LENGTH = 32
APT_SECTION_TYPE_MAX_LENGTH = 32


def np_uint16_to_string(uint16_array: np.ndarray) -> str:
    """Create string from array of uint16 numbers (UTF-16)."""
    str_parsed = ''
    for value in uint16_array:
        if value != 0:  # '\x00'
            str_parsed += chr(value)
    return f"{str_parsed}"


def string_to_typed_nparray(string: str, length: int, dtyp: type) -> np.ndarray:
    """Create length long specifically typed numpy array from string."""
    if (isinstance(dtyp, type) is True) and (len(string) <= length):
        nparr = np.zeros(length, dtype=dtyp)  # type: ignore
        for value in np.arange(0, len(string)):
            nparr[value] = ord(string[value])
        return nparr
    raise ValueError(f"{dtyp} is either not a type or {string} is not <= {length}!")



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""AMETEK APT(6) data exchange file reader used by atom probe microscopists."""

from ifes_apt_tc_data_modeling.apt.apt6_sections import AptFileSectionMetadata


# define which header and sections to expect in an *.apt file
# the sections below are referred to as branches in the commercial software
# APSuite enables users select prior I/O which of the sections to write
# normally a range file from based on the exchange with AMETEK and

# in addition there is usually a sextet of preceeding IEEE 32-bit floats
# to the position section. This sextett encodes the min,max bounds of the point
# cloud in x, y, z direction respectively
# F. M. M. de Oliveira reported cases where *.apt from mere flat test runs,
# i.e. software sessions during which no reconstruction is performed
# this sextett is absent

EXPECTED_SECTIONS = {}

# most tested sections

EXPECTED_SECTIONS["Position"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Position"].set_section_name("Position")
EXPECTED_SECTIONS["Position"].set_c_signature()
EXPECTED_SECTIONS["Position"].set_i_header_size(148 + 6 * 4)
# six 4 byte IEEE 32-bit floats are following immediately after a "Position"
# section trailing the actual position value array
EXPECTED_SECTIONS["Position"].set_i_header_version(2)
EXPECTED_SECTIONS["Position"].set_wc_section_type("Position")
EXPECTED_SECTIONS["Position"].set_i_section_version(1)
EXPECTED_SECTIONS["Position"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Position"].set_e_record_type(1)
EXPECTED_SECTIONS["Position"].set_e_record_data_type(3)
EXPECTED_SECTIONS["Position"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Position"].set_i_record_size(12)
EXPECTED_SECTIONS["Position"].set_wc_data_unit("nm")
EXPECTED_SECTIONS["Position"].set_accepted_units(["nm"])

EXPECTED_SECTIONS["Mass"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Mass"].set_section_name("Mass")
EXPECTED_SECTIONS["Mass"].set_c_signature()
EXPECTED_SECTIONS["Mass"].set_i_header_size(148)
EXPECTED_SECTIONS["Mass"].set_i_header_version(2)
EXPECTED_SECTIONS["Mass"].set_wc_section_type("Mass")
EXPECTED_SECTIONS["Mass"].set_i_section_version(1)
EXPECTED_SECTIONS["Mass"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Mass"].set_e_record_type(1)
EXPECTED_SECTIONS["Mass"].set_e_record_data_type(3)
EXPECTED_SECTIONS["Mass"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Mass"].set_i_record_size(4)
EXPECTED_SECTIONS["Mass"].set_wc_data_unit("Da")
EXPECTED_SECTIONS["Mass"].set_accepted_units(["da", "Da", "amu"])

# M. K\"uhbach detected there are at least APSuite versions which write
# files with Da and some with da as wcDataUnits argument, it seems there
# was a source code change, we use pint to detect these issues and warn

# sections which demand more testing with the parser

EXPECTED_SECTIONS["z"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["z"].set_section_name("z")
EXPECTED_SECTIONS["z"].set_c_signature()
EXPECTED_SECTIONS["z"].set_i_header_size(148)
EXPECTED_SECTIONS["z"].set_i_header_version(2)
EXPECTED_SECTIONS["z"].set_wc_section_type("z")
EXPECTED_SECTIONS["z"].set_i_section_version(1)
EXPECTED_SECTIONS["z"].set_e_relationship_type(1)
EXPECTED_SECTIONS["z"].set_e_record_type(1)
EXPECTED_SECTIONS["z"].set_e_record_data_type(1)
EXPECTED_SECTIONS["z"].set_i_data_type_size(64)
EXPECTED_SECTIONS["z"].set_i_record_size(8)
EXPECTED_SECTIONS["z"].set_wc_data_unit("ions")
EXPECTED_SECTIONS["z"].set_accepted_units(["ions"])

EXPECTED_SECTIONS["tof"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["tof"].set_section_name("tof")
EXPECTED_SECTIONS["tof"].set_c_signature()
EXPECTED_SECTIONS["tof"].set_i_header_size(148)
EXPECTED_SECTIONS["tof"].set_i_header_version(2)
EXPECTED_SECTIONS["tof"].set_wc_section_type("tof")
EXPECTED_SECTIONS["tof"].set_i_section_version(1)
EXPECTED_SECTIONS["tof"].set_e_relationship_type(1)
EXPECTED_SECTIONS["tof"].set_e_record_type(1)
EXPECTED_SECTIONS["tof"].set_e_record_data_type(3)
EXPECTED_SECTIONS["tof"].set_i_data_type_size(32)
EXPECTED_SECTIONS["tof"].set_i_record_size(4)
EXPECTED_SECTIONS["tof"].set_wc_data_unit("ns")
EXPECTED_SECTIONS["tof"].set_accepted_units(["ns"])

EXPECTED_SECTIONS["Voltage"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Voltage"].set_section_name("Voltage")
EXPECTED_SECTIONS["Voltage"].set_c_signature()
EXPECTED_SECTIONS["Voltage"].set_i_header_size(148)
EXPECTED_SECTIONS["Voltage"].set_i_header_version(2)
EXPECTED_SECTIONS["Voltage"].set_wc_section_type("Voltage")
EXPECTED_SECTIONS["Voltage"].set_i_section_version(1)
EXPECTED_SECTIONS["Voltage"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Voltage"].set_e_record_type(1)
EXPECTED_SECTIONS["Voltage"].set_e_record_data_type(3)
EXPECTED_SECTIONS["Voltage"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Voltage"].set_i_record_size(4)
EXPECTED_SECTIONS["Voltage"].set_wc_data_unit("")

EXPECTED_SECTIONS["pulse"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["pulse"].set_section_name("pulse")
EXPECTED_SECTIONS["pulse"].set_c_signature()
EXPECTED_SECTIONS["pulse"].set_i_header_size(148)
EXPECTED_SECTIONS["pulse"].set_i_header_version(2)
EXPECTED_SECTIONS["pulse"].set_wc_section_type("pulse")
EXPECTED_SECTIONS["pulse"].set_i_section_version(1)
EXPECTED_SECTIONS["pulse"].set_e_relationship_type(1)
EXPECTED_SECTIONS["pulse"].set_e_record_type(1)
EXPECTED_SECTIONS["pulse"].set_e_record_data_type(3)
EXPECTED_SECTIONS["pulse"].set_i_data_type_size(32)
EXPECTED_SECTIONS["pulse"].set_i_record_size(4)
EXPECTED_SECTIONS["pulse"].set_wc_data_unit("")

EXPECTED_SECTIONS["freq"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["freq"].set_section_name("freq")
EXPECTED_SECTIONS["freq"].set_c_signature()
EXPECTED_SECTIONS["freq"].set_i_header_size(148)
EXPECTED_SECTIONS["freq"].set_i_header_version(2)
EXPECTED_SECTIONS["freq"].set_wc_section_type("freq")
EXPECTED_SECTIONS["freq"].set_i_section_version(1)
EXPECTED_SECTIONS["freq"].set_e_relationship_type(1)
EXPECTED_SECTIONS["freq"].set_e_record_type(1)
EXPECTED_SECTIONS["freq"].set_e_record_data_type(3)
EXPECTED_SECTIONS["freq"].set_i_data_type_size(32)
EXPECTED_SECTIONS["freq"].set_i_record_size(4)
EXPECTED_SECTIONS["freq"].set_wc_data_unit("Hz")
EXPECTED_SECTIONS["freq"].set_accepted_units(["Hz"])

EXPECTED_SECTIONS["tElapsed"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["tElapsed"].set_section_name("tElapsed")
EXPECTED_SECTIONS["tElapsed"].set_c_signature()
EXPECTED_SECTIONS["tElapsed"].set_i_header_size(148)
EXPECTED_SECTIONS["tElapsed"].set_i_header_version(2)
EXPECTED_SECTIONS["tElapsed"].set_wc_section_type("tElapsed")
EXPECTED_SECTIONS["tElapsed"].set_i_section_version(1)
EXPECTED_SECTIONS["tElapsed"].set_e_relationship_type(1)
EXPECTED_SECTIONS["tElapsed"].set_e_record_type(1)
EXPECTED_SECTIONS["tElapsed"].set_e_record_data_type(3)
EXPECTED_SECTIONS["tElapsed"].set_i_data_type_size(32)
EXPECTED_SECTIONS["tElapsed"].set_i_record_size(4)
EXPECTED_SECTIONS["tElapsed"].set_wc_data_unit("")

EXPECTED_SECTIONS["erate"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["erate"].set_section_name("erate")
EXPECTED_SECTIONS["erate"].set_c_signature()
EXPECTED_SECTIONS["erate"].set_i_header_size(148)
EXPECTED_SECTIONS["erate"].set_i_header_version(2)
EXPECTED_SECTIONS["erate"].set_wc_section_type("erate")
EXPECTED_SECTIONS["erate"].set_i_section_version(1)
EXPECTED_SECTIONS["erate"].set_e_relationship_type(1)
EXPECTED_SECTIONS["erate"].set_e_record_type(1)
EXPECTED_SECTIONS["erate"].set_e_record_data_type(3)
EXPECTED_SECTIONS["erate"].set_i_data_type_size(32)
EXPECTED_SECTIONS["erate"].set_i_record_size(4)
EXPECTED_SECTIONS["erate"].set_wc_data_unit("%/100")
EXPECTED_SECTIONS["erate"].set_accepted_units(["%/100"])

EXPECTED_SECTIONS["xstage"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["xstage"].set_section_name("xstage")
EXPECTED_SECTIONS["xstage"].set_c_signature()
EXPECTED_SECTIONS["xstage"].set_i_header_size(148)
EXPECTED_SECTIONS["xstage"].set_i_header_version(2)
EXPECTED_SECTIONS["xstage"].set_wc_section_type("xstage")
EXPECTED_SECTIONS["xstage"].set_i_section_version(1)
EXPECTED_SECTIONS["xstage"].set_e_relationship_type(1)
EXPECTED_SECTIONS["xstage"].set_e_record_type(1)
EXPECTED_SECTIONS["xstage"].set_e_record_data_type(1)
EXPECTED_SECTIONS["xstage"].set_i_data_type_size(32)
EXPECTED_SECTIONS["xstage"].set_i_record_size(4)
EXPECTED_SECTIONS["xstage"].set_wc_data_unit("")

EXPECTED_SECTIONS["ystage"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["ystage"].set_section_name("ystage")
EXPECTED_SECTIONS["ystage"].set_c_signature()
EXPECTED_SECTIONS["ystage"].set_i_header_size(148)
EXPECTED_SECTIONS["ystage"].set_i_header_version(2)
EXPECTED_SECTIONS["ystage"].set_wc_section_type("ystage")
EXPECTED_SECTIONS["ystage"].set_i_section_version(1)
EXPECTED_SECTIONS["ystage"].set_e_relationship_type(1)
EXPECTED_SECTIONS["ystage"].set_e_record_type(1)
EXPECTED_SECTIONS["ystage"].set_e_record_data_type(1)
EXPECTED_SECTIONS["ystage"].set_i_data_type_size(32)
EXPECTED_SECTIONS["ystage"].set_i_record_size(4)
EXPECTED_SECTIONS["ystage"].set_wc_data_unit("")

EXPECTED_SECTIONS["zstage"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["zstage"].set_section_name("zstage")
EXPECTED_SECTIONS["zstage"].set_c_signature()
EXPECTED_SECTIONS["zstage"].set_i_header_size(148)
EXPECTED_SECTIONS["zstage"].set_i_header_version(2)
EXPECTED_SECTIONS["zstage"].set_wc_section_type("zstage")
EXPECTED_SECTIONS["zstage"].set_i_section_version(1)
EXPECTED_SECTIONS["zstage"].set_e_relationship_type(1)
EXPECTED_SECTIONS["zstage"].set_e_record_type(1)
EXPECTED_SECTIONS["zstage"].set_e_record_data_type(1)
EXPECTED_SECTIONS["zstage"].set_i_data_type_size(32)
EXPECTED_SECTIONS["zstage"].set_i_record_size(4)
EXPECTED_SECTIONS["zstage"].set_wc_data_unit("")

EXPECTED_SECTIONS["tstage"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["tstage"].set_section_name("tstage")
EXPECTED_SECTIONS["tstage"].set_c_signature()
EXPECTED_SECTIONS["tstage"].set_i_header_size(148)
EXPECTED_SECTIONS["tstage"].set_i_header_version(2)
EXPECTED_SECTIONS["tstage"].set_wc_section_type("tstage")
EXPECTED_SECTIONS["tstage"].set_i_section_version(1)
EXPECTED_SECTIONS["tstage"].set_e_relationship_type(1)
EXPECTED_SECTIONS["tstage"].set_e_record_type(1)
EXPECTED_SECTIONS["tstage"].set_e_record_data_type(2)
EXPECTED_SECTIONS["tstage"].set_i_data_type_size(16)
EXPECTED_SECTIONS["tstage"].set_i_record_size(2)
EXPECTED_SECTIONS["tstage"].set_wc_data_unit("")

EXPECTED_SECTIONS["TargetErate"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["TargetErate"].set_section_name("TargetErate")
EXPECTED_SECTIONS["TargetErate"].set_c_signature()
EXPECTED_SECTIONS["TargetErate"].set_i_header_size(148)
EXPECTED_SECTIONS["TargetErate"].set_i_header_version(2)
EXPECTED_SECTIONS["TargetErate"].set_wc_section_type("TargetErate")
EXPECTED_SECTIONS["TargetErate"].set_i_section_version(1)
EXPECTED_SECTIONS["TargetErate"].set_e_relationship_type(1)
EXPECTED_SECTIONS["TargetErate"].set_e_record_type(1)
EXPECTED_SECTIONS["TargetErate"].set_e_record_data_type(3)
EXPECTED_SECTIONS["TargetErate"].set_i_data_type_size(32)
EXPECTED_SECTIONS["TargetErate"].set_i_record_size(4)
EXPECTED_SECTIONS["TargetErate"].set_wc_data_unit("%/100")
EXPECTED_SECTIONS["TargetErate"].set_accepted_units(["%/100"])

EXPECTED_SECTIONS["TargetFlux"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["TargetFlux"].set_section_name("TargetFlux")
EXPECTED_SECTIONS["TargetFlux"].set_c_signature()
EXPECTED_SECTIONS["TargetFlux"].set_i_header_size(148)
EXPECTED_SECTIONS["TargetFlux"].set_i_header_version(2)
EXPECTED_SECTIONS["TargetFlux"].set_wc_section_type("TargetFlux")
EXPECTED_SECTIONS["TargetFlux"].set_i_section_version(1)
EXPECTED_SECTIONS["TargetFlux"].set_e_relationship_type(1)
EXPECTED_SECTIONS["TargetFlux"].set_e_record_type(1)
EXPECTED_SECTIONS["TargetFlux"].set_e_record_data_type(3)
EXPECTED_SECTIONS["TargetFlux"].set_i_data_type_size(32)
EXPECTED_SECTIONS["TargetFlux"].set_i_record_size(4)
EXPECTED_SECTIONS["TargetFlux"].set_wc_data_unit("")

EXPECTED_SECTIONS["pulseDelta"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["pulseDelta"].set_section_name("pulseDelta")
EXPECTED_SECTIONS["pulseDelta"].set_c_signature()
EXPECTED_SECTIONS["pulseDelta"].set_i_header_size(148)
EXPECTED_SECTIONS["pulseDelta"].set_i_header_version(2)
EXPECTED_SECTIONS["pulseDelta"].set_wc_section_type("pulseDelta")
EXPECTED_SECTIONS["pulseDelta"].set_i_section_version(1)
EXPECTED_SECTIONS["pulseDelta"].set_e_relationship_type(1)
EXPECTED_SECTIONS["pulseDelta"].set_e_record_type(1)
EXPECTED_SECTIONS["pulseDelta"].set_e_record_data_type(1)
EXPECTED_SECTIONS["pulseDelta"].set_i_data_type_size(16)
EXPECTED_SECTIONS["pulseDelta"].set_i_record_size(2)
EXPECTED_SECTIONS["pulseDelta"].set_wc_data_unit("")

EXPECTED_SECTIONS["Pres"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Pres"].set_section_name("Pres")
EXPECTED_SECTIONS["Pres"].set_c_signature()
EXPECTED_SECTIONS["Pres"].set_i_header_size(148)
EXPECTED_SECTIONS["Pres"].set_i_header_version(2)
EXPECTED_SECTIONS["Pres"].set_wc_section_type("Pres")
EXPECTED_SECTIONS["Pres"].set_i_section_version(1)
EXPECTED_SECTIONS["Pres"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Pres"].set_e_record_type(1)
EXPECTED_SECTIONS["Pres"].set_e_record_data_type(3)
EXPECTED_SECTIONS["Pres"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Pres"].set_i_record_size(4)
EXPECTED_SECTIONS["Pres"].set_wc_data_unit("torr")
EXPECTED_SECTIONS["Pres"].set_accepted_units(["torr"])

EXPECTED_SECTIONS["VAnodeMon"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["VAnodeMon"].set_section_name("VAnodeMon")
EXPECTED_SECTIONS["VAnodeMon"].set_c_signature()
EXPECTED_SECTIONS["VAnodeMon"].set_i_header_size(148)
EXPECTED_SECTIONS["VAnodeMon"].set_i_header_version(2)
EXPECTED_SECTIONS["VAnodeMon"].set_wc_section_type("VAnodeMon")
EXPECTED_SECTIONS["VAnodeMon"].set_i_section_version(1)
EXPECTED_SECTIONS["VAnodeMon"].set_e_relationship_type(1)
EXPECTED_SECTIONS["VAnodeMon"].set_e_record_type(1)
EXPECTED_SECTIONS["VAnodeMon"].set_e_record_data_type(3)
EXPECTED_SECTIONS["VAnodeMon"].set_i_data_type_size(32)
EXPECTED_SECTIONS["VAnodeMon"].set_i_record_size(4)
EXPECTED_SECTIONS["VAnodeMon"].set_wc_data_unit("")

EXPECTED_SECTIONS["Temp"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Temp"].set_section_name("Temp")
EXPECTED_SECTIONS["Temp"].set_c_signature()
EXPECTED_SECTIONS["Temp"].set_i_header_size(148)
EXPECTED_SECTIONS["Temp"].set_i_header_version(2)
EXPECTED_SECTIONS["Temp"].set_wc_section_type("Temp")
EXPECTED_SECTIONS["Temp"].set_i_section_version(1)
EXPECTED_SECTIONS["Temp"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Temp"].set_e_record_type(1)
EXPECTED_SECTIONS["Temp"].set_e_record_data_type(3)
EXPECTED_SECTIONS["Temp"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Temp"].set_i_record_size(4)
EXPECTED_SECTIONS["Temp"].set_wc_data_unit("K")
EXPECTED_SECTIONS["Temp"].set_accepted_units(["K"])


EXPECTED_SECTIONS["AmbTemp"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["AmbTemp"].set_section_name("AmbTemp")
EXPECTED_SECTIONS["AmbTemp"].set_c_signature()
EXPECTED_SECTIONS["AmbTemp"].set_i_header_size(148)
EXPECTED_SECTIONS["AmbTemp"].set_i_header_version(2)
EXPECTED_SECTIONS["AmbTemp"].set_wc_section_type("AmbTemp")
EXPECTED_SECTIONS["AmbTemp"].set_i_section_version(1)
EXPECTED_SECTIONS["AmbTemp"].set_e_relationship_type(1)
EXPECTED_SECTIONS["AmbTemp"].set_e_record_type(1)
EXPECTED_SECTIONS["AmbTemp"].set_e_record_data_type(3)
EXPECTED_SECTIONS["AmbTemp"].set_i_data_type_size(32)
EXPECTED_SECTIONS["AmbTemp"].set_i_record_size(4)
EXPECTED_SECTIONS["AmbTemp"].set_wc_data_unit("C")
EXPECTED_SECTIONS["AmbTemp"].set_accepted_units(["C"])

EXPECTED_SECTIONS["laserx"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["laserx"].set_section_name("laserx")
EXPECTED_SECTIONS["laserx"].set_c_signature()
EXPECTED_SECTIONS["laserx"].set_i_header_size(148)
EXPECTED_SECTIONS["laserx"].set_i_header_version(2)
EXPECTED_SECTIONS["laserx"].set_wc_section_type("laserx")
EXPECTED_SECTIONS["laserx"].set_i_section_version(1)
EXPECTED_SECTIONS["laserx"].set_e_relationship_type(1)
EXPECTED_SECTIONS["laserx"].set_e_record_type(1)
EXPECTED_SECTIONS["laserx"].set_e_record_data_type(1)
EXPECTED_SECTIONS["laserx"].set_i_data_type_size(32)
EXPECTED_SECTIONS["laserx"].set_i_record_size(4)
EXPECTED_SECTIONS["laserx"].set_wc_data_unit("")

EXPECTED_SECTIONS["lasery"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["lasery"].set_section_name("lasery")
EXPECTED_SECTIONS["lasery"].set_c_signature()
EXPECTED_SECTIONS["lasery"].set_i_header_size(148)
EXPECTED_SECTIONS["lasery"].set_i_header_version(2)
EXPECTED_SECTIONS["lasery"].set_wc_section_type("lasery")
EXPECTED_SECTIONS["lasery"].set_i_section_version(1)
EXPECTED_SECTIONS["lasery"].set_e_relationship_type(1)
EXPECTED_SECTIONS["lasery"].set_e_record_type(1)
EXPECTED_SECTIONS["lasery"].set_e_record_data_type(1)
EXPECTED_SECTIONS["lasery"].set_i_data_type_size(32)
EXPECTED_SECTIONS["lasery"].set_i_record_size(4)
EXPECTED_SECTIONS["lasery"].set_wc_data_unit("")

EXPECTED_SECTIONS["laserz"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["laserz"].set_section_name("laserz")
EXPECTED_SECTIONS["laserz"].set_c_signature()
EXPECTED_SECTIONS["laserz"].set_i_header_size(148)
EXPECTED_SECTIONS["laserz"].set_i_header_version(2)
EXPECTED_SECTIONS["laserz"].set_wc_section_type("laserz")
EXPECTED_SECTIONS["laserz"].set_i_section_version(1)
EXPECTED_SECTIONS["laserz"].set_e_relationship_type(1)
EXPECTED_SECTIONS["laserz"].set_e_record_type(1)
EXPECTED_SECTIONS["laserz"].set_e_record_data_type(1)
EXPECTED_SECTIONS["laserz"].set_i_data_type_size(32)
EXPECTED_SECTIONS["laserz"].set_i_record_size(4)
EXPECTED_SECTIONS["laserz"].set_wc_data_unit("")

EXPECTED_SECTIONS["laserpower"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["laserpower"].set_section_name("laserpower")
EXPECTED_SECTIONS["laserpower"].set_c_signature()
EXPECTED_SECTIONS["laserpower"].set_i_header_size(148)
EXPECTED_SECTIONS["laserpower"].set_i_header_version(2)
EXPECTED_SECTIONS["laserpower"].set_wc_section_type("laserpower")
EXPECTED_SECTIONS["laserpower"].set_i_section_version(1)
EXPECTED_SECTIONS["laserpower"].set_e_relationship_type(1)
EXPECTED_SECTIONS["laserpower"].set_e_record_type(1)
EXPECTED_SECTIONS["laserpower"].set_e_record_data_type(3)
EXPECTED_SECTIONS["laserpower"].set_i_data_type_size(32)
EXPECTED_SECTIONS["laserpower"].set_i_record_size(4)
EXPECTED_SECTIONS["laserpower"].set_wc_data_unit("")

EXPECTED_SECTIONS["FractureGuard"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["FractureGuard"].set_section_name("FractureGuard")
EXPECTED_SECTIONS["FractureGuard"].set_c_signature()
EXPECTED_SECTIONS["FractureGuard"].set_i_header_size(148)
EXPECTED_SECTIONS["FractureGuard"].set_i_header_version(2)
EXPECTED_SECTIONS["FractureGuard"].set_wc_section_type("FractureGuard")
EXPECTED_SECTIONS["FractureGuard"].set_i_section_version(1)
EXPECTED_SECTIONS["FractureGuard"].set_e_relationship_type(1)
EXPECTED_SECTIONS["FractureGuard"].set_e_record_type(1)
EXPECTED_SECTIONS["FractureGuard"].set_e_record_data_type(2)
EXPECTED_SECTIONS["FractureGuard"].set_i_data_type_size(16)
EXPECTED_SECTIONS["FractureGuard"].set_i_record_size(2)
EXPECTED_SECTIONS["FractureGuard"].set_wc_data_unit("")

EXPECTED_SECTIONS["Noise"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Noise"].set_section_name("Noise")
EXPECTED_SECTIONS["Noise"].set_c_signature()
EXPECTED_SECTIONS["Noise"].set_i_header_size(148)
EXPECTED_SECTIONS["Noise"].set_i_header_version(2)
EXPECTED_SECTIONS["Noise"].set_wc_section_type("Noise")
EXPECTED_SECTIONS["Noise"].set_i_section_version(1)
EXPECTED_SECTIONS["Noise"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Noise"].set_e_record_type(1)
EXPECTED_SECTIONS["Noise"].set_e_record_data_type(3)
EXPECTED_SECTIONS["Noise"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Noise"].set_i_record_size(4)
EXPECTED_SECTIONS["Noise"].set_wc_data_unit("ions")
EXPECTED_SECTIONS["Noise"].set_accepted_units(["", "ions"])

EXPECTED_SECTIONS["Uniformity"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Uniformity"].set_section_name("Uniformity")
EXPECTED_SECTIONS["Uniformity"].set_c_signature()
EXPECTED_SECTIONS["Uniformity"].set_i_header_size(148)
EXPECTED_SECTIONS["Uniformity"].set_i_header_version(2)
EXPECTED_SECTIONS["Uniformity"].set_wc_section_type("Uniformity")
EXPECTED_SECTIONS["Uniformity"].set_i_section_version(1)
EXPECTED_SECTIONS["Uniformity"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Uniformity"].set_e_record_type(1)
EXPECTED_SECTIONS["Uniformity"].set_e_record_data_type(3)
EXPECTED_SECTIONS["Uniformity"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Uniformity"].set_i_record_size(4)
EXPECTED_SECTIONS["Uniformity"].set_wc_data_unit("")

EXPECTED_SECTIONS["tofc"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["tofc"].set_section_name("tofc")
EXPECTED_SECTIONS["tofc"].set_c_signature()
EXPECTED_SECTIONS["tofc"].set_i_header_size(148)
EXPECTED_SECTIONS["tofc"].set_i_header_version(2)
EXPECTED_SECTIONS["tofc"].set_wc_section_type("tofc")
EXPECTED_SECTIONS["tofc"].set_i_section_version(1)
EXPECTED_SECTIONS["tofc"].set_e_relationship_type(1)
EXPECTED_SECTIONS["tofc"].set_e_record_type(1)
EXPECTED_SECTIONS["tofc"].set_e_record_data_type(3)
EXPECTED_SECTIONS["tofc"].set_i_data_type_size(32)
EXPECTED_SECTIONS["tofc"].set_i_record_size(4)
EXPECTED_SECTIONS["tofc"].set_wc_data_unit("ns")
EXPECTED_SECTIONS["tofc"].set_accepted_units(["ns"])

EXPECTED_SECTIONS["tofb"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["tofb"].set_section_name("tofb")
EXPECTED_SECTIONS["tofb"].set_c_signature()
EXPECTED_SECTIONS["tofb"].set_i_header_size(148)
EXPECTED_SECTIONS["tofb"].set_i_header_version(2)
EXPECTED_SECTIONS["tofb"].set_wc_section_type("tofb")
EXPECTED_SECTIONS["tofb"].set_i_section_version(1)
EXPECTED_SECTIONS["tofb"].set_e_relationship_type(1)
EXPECTED_SECTIONS["tofb"].set_e_record_type(1)
EXPECTED_SECTIONS["tofb"].set_e_record_data_type(3)
EXPECTED_SECTIONS["tofb"].set_i_data_type_size(32)
EXPECTED_SECTIONS["tofb"].set_i_record_size(4)
EXPECTED_SECTIONS["tofb"].set_wc_data_unit("ns")
EXPECTED_SECTIONS["tofb"].set_accepted_units(["ns"])

EXPECTED_SECTIONS["xs"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["xs"].set_section_name("xs")
EXPECTED_SECTIONS["xs"].set_c_signature()
EXPECTED_SECTIONS["xs"].set_i_header_size(148)
EXPECTED_SECTIONS["xs"].set_i_header_version(2)
EXPECTED_SECTIONS["xs"].set_wc_section_type("xs")
EXPECTED_SECTIONS["xs"].set_i_section_version(1)
EXPECTED_SECTIONS["xs"].set_e_relationship_type(1)
EXPECTED_SECTIONS["xs"].set_e_record_type(1)
EXPECTED_SECTIONS["xs"].set_e_record_data_type(3)
EXPECTED_SECTIONS["xs"].set_i_data_type_size(32)
EXPECTED_SECTIONS["xs"].set_i_record_size(4)
EXPECTED_SECTIONS["xs"].set_wc_data_unit("nm")
EXPECTED_SECTIONS["xs"].set_accepted_units(["nm"])

EXPECTED_SECTIONS["ys"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["ys"].set_section_name("ys")
EXPECTED_SECTIONS["ys"].set_c_signature()
EXPECTED_SECTIONS["ys"].set_i_header_size(148)
EXPECTED_SECTIONS["ys"].set_i_header_version(2)
EXPECTED_SECTIONS["ys"].set_wc_section_type("ys")
EXPECTED_SECTIONS["ys"].set_i_section_version(1)
EXPECTED_SECTIONS["ys"].set_e_relationship_type(1)
EXPECTED_SECTIONS["ys"].set_e_record_type(1)
EXPECTED_SECTIONS["ys"].set_e_record_data_type(3)
EXPECTED_SECTIONS["ys"].set_i_data_type_size(32)
EXPECTED_SECTIONS["ys"].set_i_record_size(4)
EXPECTED_SECTIONS["ys"].set_wc_data_unit("nm")
EXPECTED_SECTIONS["ys"].set_accepted_units(["nm"])

EXPECTED_SECTIONS["zs"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["zs"].set_section_name("zs")
EXPECTED_SECTIONS["zs"].set_c_signature()
EXPECTED_SECTIONS["zs"].set_i_header_size(148)
EXPECTED_SECTIONS["zs"].set_i_header_version(2)
EXPECTED_SECTIONS["zs"].set_wc_section_type("zs")
EXPECTED_SECTIONS["zs"].set_i_section_version(1)
EXPECTED_SECTIONS["zs"].set_e_relationship_type(1)
EXPECTED_SECTIONS["zs"].set_e_record_type(1)
EXPECTED_SECTIONS["zs"].set_e_record_data_type(3)
EXPECTED_SECTIONS["zs"].set_i_data_type_size(32)
EXPECTED_SECTIONS["zs"].set_i_record_size(4)
EXPECTED_SECTIONS["zs"].set_wc_data_unit("nm")
EXPECTED_SECTIONS["zs"].set_accepted_units(["nm"])

EXPECTED_SECTIONS["rTip"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["rTip"].set_section_name("rTip")
EXPECTED_SECTIONS["rTip"].set_c_signature()
EXPECTED_SECTIONS["rTip"].set_i_header_size(148)
EXPECTED_SECTIONS["rTip"].set_i_header_version(2)
EXPECTED_SECTIONS["rTip"].set_wc_section_type("rTip")
EXPECTED_SECTIONS["rTip"].set_i_section_version(1)
EXPECTED_SECTIONS["rTip"].set_e_relationship_type(1)
EXPECTED_SECTIONS["rTip"].set_e_record_type(1)
EXPECTED_SECTIONS["rTip"].set_e_record_data_type(3)
EXPECTED_SECTIONS["rTip"].set_i_data_type_size(32)
EXPECTED_SECTIONS["rTip"].set_i_record_size(4)
EXPECTED_SECTIONS["rTip"].set_wc_data_unit("nm")
EXPECTED_SECTIONS["rTip"].set_accepted_units(["nm"])

EXPECTED_SECTIONS["zApex"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["zApex"].set_section_name("zApex")
EXPECTED_SECTIONS["zApex"].set_c_signature()
EXPECTED_SECTIONS["zApex"].set_i_header_size(148)
EXPECTED_SECTIONS["zApex"].set_i_header_version(2)
EXPECTED_SECTIONS["zApex"].set_wc_section_type("zApex")
EXPECTED_SECTIONS["zApex"].set_i_section_version(1)
EXPECTED_SECTIONS["zApex"].set_e_relationship_type(1)
EXPECTED_SECTIONS["zApex"].set_e_record_type(1)
EXPECTED_SECTIONS["zApex"].set_e_record_data_type(3)
EXPECTED_SECTIONS["zApex"].set_i_data_type_size(32)
EXPECTED_SECTIONS["zApex"].set_i_record_size(4)
EXPECTED_SECTIONS["zApex"].set_wc_data_unit("nm")
EXPECTED_SECTIONS["zApex"].set_accepted_units(["nm"])

EXPECTED_SECTIONS["zSphereCorr"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["zSphereCorr"].set_section_name("zSphereCorr")
EXPECTED_SECTIONS["zSphereCorr"].set_c_signature()
EXPECTED_SECTIONS["zSphereCorr"].set_i_header_size(148)
EXPECTED_SECTIONS["zSphereCorr"].set_i_header_version(2)
EXPECTED_SECTIONS["zSphereCorr"].set_wc_section_type("zSphereCorr")
EXPECTED_SECTIONS["zSphereCorr"].set_i_section_version(1)
EXPECTED_SECTIONS["zSphereCorr"].set_e_relationship_type(1)
EXPECTED_SECTIONS["zSphereCorr"].set_e_record_type(1)
EXPECTED_SECTIONS["zSphereCorr"].set_e_record_data_type(3)
EXPECTED_SECTIONS["zSphereCorr"].set_i_data_type_size(32)
EXPECTED_SECTIONS["zSphereCorr"].set_i_record_size(4)
EXPECTED_SECTIONS["zSphereCorr"].set_wc_data_unit("nm")
EXPECTED_SECTIONS["zSphereCorr"].set_accepted_units(["nm"])

# the next three seem to (have been /were used for AMETEK development purposes
EXPECTED_SECTIONS["Position_0"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Position_0"].set_section_name("Position_0")
EXPECTED_SECTIONS["Position_0"].set_c_signature()
EXPECTED_SECTIONS["Position_0"].set_i_header_size(148 + 6 * 4)
EXPECTED_SECTIONS["Position_0"].set_i_header_version(2)
EXPECTED_SECTIONS["Position_0"].set_wc_section_type("Position_0")
EXPECTED_SECTIONS["Position_0"].set_i_section_version(1)
EXPECTED_SECTIONS["Position_0"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Position_0"].set_e_record_type(1)
EXPECTED_SECTIONS["Position_0"].set_e_record_data_type(3)
EXPECTED_SECTIONS["Position_0"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Position_0"].set_i_record_size(12)
EXPECTED_SECTIONS["Position_0"].set_wc_data_unit("")

EXPECTED_SECTIONS["Position_1"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Position_1"].set_section_name("Position_1")
EXPECTED_SECTIONS["Position_1"].set_c_signature()
EXPECTED_SECTIONS["Position_1"].set_i_header_size(148 + 6 * 4)
EXPECTED_SECTIONS["Position_1"].set_i_header_version(2)
EXPECTED_SECTIONS["Position_1"].set_wc_section_type("Position_1")
EXPECTED_SECTIONS["Position_1"].set_i_section_version(1)
EXPECTED_SECTIONS["Position_1"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Position_1"].set_e_record_type(1)
EXPECTED_SECTIONS["Position_1"].set_e_record_data_type(3)
EXPECTED_SECTIONS["Position_1"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Position_1"].set_i_record_size(12)
EXPECTED_SECTIONS["Position_1"].set_wc_data_unit("")

EXPECTED_SECTIONS["Position_2"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Position_2"].set_section_name("Position_2")
EXPECTED_SECTIONS["Position_2"].set_c_signature()
EXPECTED_SECTIONS["Position_2"].set_i_header_size(148 + 6 * 4)
EXPECTED_SECTIONS["Position_2"].set_i_header_version(2)
EXPECTED_SECTIONS["Position_2"].set_wc_section_type("Position_2")
EXPECTED_SECTIONS["Position_2"].set_i_section_version(1)
EXPECTED_SECTIONS["Position_2"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Position_2"].set_e_record_type(1)
EXPECTED_SECTIONS["Position_2"].set_e_record_data_type(3)
EXPECTED_SECTIONS["Position_2"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Position_2"].set_i_record_size(12)
EXPECTED_SECTIONS["Position_2"].set_wc_data_unit("")

EXPECTED_SECTIONS["XDet_mm"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["XDet_mm"].set_section_name("XDet_mm")
EXPECTED_SECTIONS["XDet_mm"].set_c_signature()
EXPECTED_SECTIONS["XDet_mm"].set_i_header_size(148)
EXPECTED_SECTIONS["XDet_mm"].set_i_header_version(2)
EXPECTED_SECTIONS["XDet_mm"].set_wc_section_type("XDet_mm")
EXPECTED_SECTIONS["XDet_mm"].set_i_section_version(1)
EXPECTED_SECTIONS["XDet_mm"].set_e_relationship_type(1)
EXPECTED_SECTIONS["XDet_mm"].set_e_record_type(1)
EXPECTED_SECTIONS["XDet_mm"].set_e_record_data_type(3)
EXPECTED_SECTIONS["XDet_mm"].set_i_data_type_size(32)
EXPECTED_SECTIONS["XDet_mm"].set_i_record_size(4)
EXPECTED_SECTIONS["XDet_mm"].set_wc_data_unit("mm")
EXPECTED_SECTIONS["XDet_mm"].set_accepted_units(["mm"])

EXPECTED_SECTIONS["YDet_mm"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["YDet_mm"].set_section_name("YDet_mm")
EXPECTED_SECTIONS["YDet_mm"].set_c_signature()
EXPECTED_SECTIONS["YDet_mm"].set_i_header_size(148)
EXPECTED_SECTIONS["YDet_mm"].set_i_header_version(2)
EXPECTED_SECTIONS["YDet_mm"].set_wc_section_type("YDet_mm")
EXPECTED_SECTIONS["YDet_mm"].set_i_section_version(1)
EXPECTED_SECTIONS["YDet_mm"].set_e_relationship_type(1)
EXPECTED_SECTIONS["YDet_mm"].set_e_record_type(1)
EXPECTED_SECTIONS["YDet_mm"].set_e_record_data_type(3)
EXPECTED_SECTIONS["YDet_mm"].set_i_data_type_size(32)
EXPECTED_SECTIONS["YDet_mm"].set_i_record_size(4)
EXPECTED_SECTIONS["YDet_mm"].set_wc_data_unit("mm")
EXPECTED_SECTIONS["YDet_mm"].set_accepted_units(["mm"])

EXPECTED_SECTIONS["Multiplicity"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Multiplicity"].set_section_name("Multiplicity")
EXPECTED_SECTIONS["Multiplicity"].set_c_signature()
EXPECTED_SECTIONS["Multiplicity"].set_i_header_size(148)
EXPECTED_SECTIONS["Multiplicity"].set_i_header_version(2)
EXPECTED_SECTIONS["Multiplicity"].set_wc_section_type("Multiplicity")
EXPECTED_SECTIONS["Multiplicity"].set_i_section_version(1)
EXPECTED_SECTIONS["Multiplicity"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Multiplicity"].set_e_record_type(1)
EXPECTED_SECTIONS["Multiplicity"].set_e_record_data_type(1)
EXPECTED_SECTIONS["Multiplicity"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Multiplicity"].set_i_record_size(4)
EXPECTED_SECTIONS["Multiplicity"].set_wc_data_unit("")

EXPECTED_SECTIONS["Vap"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Vap"].set_section_name("Vap")
EXPECTED_SECTIONS["Vap"].set_c_signature()
EXPECTED_SECTIONS["Vap"].set_i_header_size(148)
EXPECTED_SECTIONS["Vap"].set_i_header_version(2)
EXPECTED_SECTIONS["Vap"].set_wc_section_type("Vap")
EXPECTED_SECTIONS["Vap"].set_i_section_version(1)
EXPECTED_SECTIONS["Vap"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Vap"].set_e_record_type(1)
EXPECTED_SECTIONS["Vap"].set_e_record_data_type(3)
EXPECTED_SECTIONS["Vap"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Vap"].set_i_record_size(4)
EXPECTED_SECTIONS["Vap"].set_wc_data_unit("V")
EXPECTED_SECTIONS["Vap"].set_accepted_units(["V"])

EXPECTED_SECTIONS["Detector Coordinates"] = AptFileSectionMetadata()
EXPECTED_SECTIONS["Detector Coordinates"].set_section_name(
    "Detector Coordinates")
EXPECTED_SECTIONS["Detector Coordinates"].set_c_signature()
EXPECTED_SECTIONS["Detector Coordinates"].set_i_header_size(148)
EXPECTED_SECTIONS["Detector Coordinates"].set_i_header_version(2)
EXPECTED_SECTIONS["Detector Coordinates"].set_wc_section_type(
    "Detector Coordinates")
EXPECTED_SECTIONS["Detector Coordinates"].set_i_section_version(1)
EXPECTED_SECTIONS["Detector Coordinates"].set_e_relationship_type(1)
EXPECTED_SECTIONS["Detector Coordinates"].set_e_record_type(1)
EXPECTED_SECTIONS["Detector Coordinates"].set_e_record_data_type(3)
EXPECTED_SECTIONS["Detector Coordinates"].set_i_data_type_size(32)
EXPECTED_SECTIONS["Detector Coordinates"].set_i_record_size(8)
EXPECTED_SECTIONS["Detector Coordinates"].set_wc_data_unit("mm")
EXPECTED_SECTIONS["Detector Coordinates"].set_accepted_units(["mm"])

# there is at least a hint that some time ago this section"s wcDataUnit was ""


# sections whose formatting and purpose is completely unclear

# Var44 ? magic-in-action? M. K\"uhbach, could well for AMETEK development only
# EXPECTED_SECTIONS["Var44"].set_section_name("Var44")


# deprecated sections or sections with detected inconsistencies across versions
# Vref vs Voltage branch issue
EXPECTED_SECTIONS["Vref"] = EXPECTED_SECTIONS["Voltage"]
EXPECTED_SECTIONS["Vref"].set_wc_data_unit("V")
EXPECTED_SECTIONS["Vref"].set_accepted_units(["V"])

# pulseDelta vs Delta Pulse issue
# at least in one case a section Delta Pulse appeared
# at least in one case a section Epos ToF appeared

# other comments and issues
# Need to check APSuite version and build number
# "Voltage" section, M. K\"uhbach: expect to have units but example *.apt files
# from flat test do not encode a unit in the "Voltage" section



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Set of utility tools for parsing file formats used by atom probe."""

# pylint: disable=too-many-instance-attributes,unused-variable

import numpy as np

from ifes_apt_tc_data_modeling.utils.definitions import \
    MAX_NUMBER_OF_ATOMS_PER_ION
from ifes_apt_tc_data_modeling.utils.utils import \
    create_nuclide_hash, nuclide_hash_to_nuclide_list, \
    nuclide_hash_to_human_readable_name, is_range_significant
from ifes_apt_tc_data_modeling.utils.molecular_ions import \
    MolecularIonCandidate, MolecularIonBuilder, \
    PRACTICAL_ABUNDANCE, PRACTICAL_ABUNDANCE_PRODUCT, \
    PRACTICAL_MIN_HALF_LIFE, VERBOSE, SACRIFICE_ISOTOPIC_UNIQUENESS
from ifes_apt_tc_data_modeling.nexus.nx_field import NxField


def try_to_reduce_to_unique_definitions(inp: list) -> list:
    """Try to reduce a set of (molecular) ion definitions to unique."""
    for entry in inp:
        if isinstance(entry, NxIon):
            continue
        else:
            raise ValueError(f"Argument inp to try_to_reduce_to_unique_definitions needs to list of NxIon!")
    unique = []
    # unique if mqival does not overlap (but can touch) either side
    # extrema of ranging definition and ivec is different or all 0
    # from a scientific point of view we would like iontypes to be
    # unique and ranges at most touching numerically but not overlapping
    # as then for a given mass-to-charge-state value an ion can qualify
    # to be an instance more than one iontype thus making the ranging
    # ambiguous
    visited = np.asarray(np.zeros(len(inp,)), bool)
    for idx in np.arange(0, len(inp)):
        if not visited[idx]:
            # find all ranging definition value intersections with other ions
            isect = []  # 
            for jdx in np.concatenate((np.arange(0, idx), np.arange(idx + 1, len(inp)))):
                if not visited[jdx]:
                    if inp[idx].ranges.values[0, 1] < inp[jdx].ranges.values[0, 0] \
                        or inp[idx].ranges.values[0, 0] > inp[jdx].ranges.values[0, 1]:
                        continue
                    else:
                        # append only if exactly the same ivec
                        idx_jdx_are_equal = True  # try to falsify
                        for i in np.arange(0, MAX_NUMBER_OF_ATOMS_PER_ION):
                            if inp[idx].nuclide_hash.values[i] != inp[jdx].nuclide_hash.values[i]:
                                idx_jdx_are_equal = False
                                break
                        if idx_jdx_are_equal:
                            isect.append(jdx)
                            # that nuclide_hashes are the same is necessary for subsequent
                            # processing of the range for these ions
                        """
                        else:
                            print(f"Overlapping or exactly numerically aligned ranges for different ion types {idx}, {jdx}!")
                            inp[idx].report()
                            inp[jdx].report()
                        """
            # print(f"isect {isect}")
            # if there are none accept this candidate for sure
            visited[idx] = True
            if len(isect) == 0:
                # inp[idx].report()
                unique.append(inp[idx])
            else:
                # combine range of isect candidates with the same nuclide_hash
                mqmin = inp[idx].ranges.values[0, 0]
                mqmax = inp[idx].ranges.values[0, 1]
                for ids in isect:
                    visited[ids] = True
                    if inp[ids].ranges.values[0, 0] <= mqmin:
                        mqmin = inp[ids].ranges.values[0, 0]
                    if inp[ids].ranges.values[0, 1] >= mqmax:
                        mqmax = inp[ids].ranges.values[0, 1]
                joined_ion = NxIon(nuclide_hash=inp[idx].nuclide_hash.values, charge_state=0)
                joined_ion.add_range(mqmin, mqmax)
                joined_ion.comment.values = f"{inp[idx].comment.values} was combined with {isect}"
                # joined_ion.report()
                unique.append(joined_ion)
    return unique


class NxIon():
    """Representative of a NeXus base class NXion."""

    def __init__(self, *args, **kwargs):
        self.comment = NxField("", "")  # comment, use e.g. for label of custom ion types
        self.color = NxField("", "")  # color used by software which created the dataset
        self.volume = NxField("", "")  # volume value in range files
        self.ion_type = NxField("", "")
        self.charge_state_model = {}
        if len(args) >= 1:
            if isinstance(args[0], list) is False:
                raise ValueError("args[0] needs to be a list!")
            self.nuclide_hash = NxField(create_nuclide_hash(args[0]), "")
        elif "nuclide_hash" in kwargs:
            if isinstance(kwargs["nuclide_hash"], np.ndarray) is False:
                raise ValueError("kwargs nuclide_hash needs to be an np.ndarray!")
            if np.shape(kwargs["nuclide_hash"]) != (MAX_NUMBER_OF_ATOMS_PER_ION,):
                raise ValueError(
                    f"kwargs nuclide_hash needs be a ({MAX_NUMBER_OF_ATOMS_PER_ION},) array!")
            self.nuclide_hash = NxField(np.asarray(kwargs["nuclide_hash"], np.uint16), "")
        else:
            # the default UNKNOWN IONTYPE
            self.nuclide_hash = NxField(create_nuclide_hash([]), "")
        self.nuclide_list = NxField(nuclide_hash_to_nuclide_list(self.nuclide_hash.values), "")
        if "charge_state" in kwargs:
            if isinstance(kwargs["charge_state"], int) \
                    and (-8 < kwargs["charge_state"] < +8):
                self.charge_state = NxField(np.int8(kwargs["charge_state"]), "")
        else:
            # charge_state 0 flags and warns that it was impossible to recover
            # the relevant charge which is usually a sign that the range
            # is not matching the theoretically expect peak location
            self.charge_state = NxField(np.int8(0), "")
        self.name = NxField(nuclide_hash_to_human_readable_name(
            self.nuclide_hash.values, self.charge_state.values))
        self.ranges = NxField(np.empty((0, 2), np.float64), "amu")

    def add_range(self, mqmin: np.float64, mqmax: np.float64):
        """Adding mass-to-charge-state ratio interval."""
        if is_range_significant(mqmin, mqmax) is False:
            raise ValueError(f"Refusing to add epsilon range [{mqmin}, {mqmax}] !")
        # the following example shows that is_range_overlapping should not be checked for
        # like it was in the past
        # ion.add_range(10.0, 12.0), ion.add_range(12.0, 13.3)
        # is equivalent to ion.add_range(10.0, 13.3)
        # assert is_range_overlapping(np.asarray([mqmin, mqmax]),
        #                            self.ranges.values) is False, \
        #                            "Refusing overlapping range!"
        self.ranges.values = np.vstack((self.ranges.values, np.array([mqmin, mqmax])))

    def update_human_readable_name(self):
        """Re-evaluate charge and nuclide_hash for name."""
        self.name = NxField(nuclide_hash_to_human_readable_name(
            self.nuclide_hash.values, self.charge_state.values))

    def report(self):
        """Report values."""
        print(f"ion_type: {self.ion_type.values}\n"
              f"nuclide_hash: {self.nuclide_hash.values}\n"
              f"nuclide_list: {self.nuclide_list.values}\n"
              f"human-readable name: {self.name.values}\n"
              f"charge_state: {self.charge_state.values}\n"
              f"ranges: {self.ranges.values}\n"
              f"comment: {self.comment.values}\n"
              f"color: {self.color.values}\n"
              f"volume: {self.volume.values}\n")

    def apply_combinatorics(self):
        """Apply specifically constrainted combinatorial analysis."""
        crawler = MolecularIonBuilder(
            min_abundance=PRACTICAL_ABUNDANCE,
            min_abundance_product=PRACTICAL_ABUNDANCE_PRODUCT,
            min_half_life=PRACTICAL_MIN_HALF_LIFE,
            sacrifice_uniqueness=SACRIFICE_ISOTOPIC_UNIQUENESS,
            verbose=VERBOSE)
        recovered_charge_state, m_ion_candidates = crawler.combinatorics(
            self.nuclide_hash.values,
            self.ranges.values[0, 0],
            self.ranges.values[0, 1])
        # print(f"{recovered_charge_state}")
        self.charge_state = NxField(np.int8(recovered_charge_state), "")
        self.update_human_readable_name()
        self.add_charge_state_model({"min_abundance": PRACTICAL_ABUNDANCE,
                                     "min_abundance_product": PRACTICAL_ABUNDANCE_PRODUCT,
                                     "min_half_life": PRACTICAL_MIN_HALF_LIFE,
                                     "sacrifice_isotopic_uniqueness": SACRIFICE_ISOTOPIC_UNIQUENESS},
                                    crawler.candidates)

    def add_charge_state_model(self,
                               parameters,
                               candidates):
        """Add details about the model how self.charge_state was defined."""
        self.charge_state_model = {}
        req_parms = ["min_abundance", "min_abundance_product",
                     "min_half_life", "sacrifice_isotopic_uniqueness"]
        for req in req_parms:
            if req in parameters:
                continue
            raise ValueError(f"Parameter {req} not defined in parameters dict!")
        self.charge_state_model = {"n_cand": 0}
        for key, val in parameters.items():
            if key not in self.charge_state_model:
                self.charge_state_model[key] = val
        n_cand = 0
        for cand in candidates:
            if isinstance(cand, MolecularIonCandidate):
                n_cand += 1
        if n_cand == 0:
            return
        self.charge_state_model["n_cand"] = n_cand
        if n_cand == 1:
            self.charge_state_model["nuclide_hash"] \
                = np.zeros((1, MAX_NUMBER_OF_ATOMS_PER_ION), np.uint16)
            self.charge_state_model["nuclide_hash"][0, 0:len(candidates[0].nuclide_hash)] \
                = candidates[0].nuclide_hash
            self.charge_state_model["charge_state"] = candidates[0].charge_state
            self.charge_state_model["mass"] = candidates[0].mass
            self.charge_state_model["natural_abundance_product"] = candidates[0].abundance_product
            self.charge_state_model["shortest_half_life"] = candidates[0].shortest_half_life
        else:
            self.charge_state_model["nuclide_hash"] \
                = np.zeros((n_cand, MAX_NUMBER_OF_ATOMS_PER_ION), np.uint16)
            self.charge_state_model["charge_state"] = np.zeros((n_cand,), np.int8)
            self.charge_state_model["mass"] = np.zeros((n_cand,), np.float64)
            self.charge_state_model["natural_abundance_product"] = np.zeros((n_cand,), np.float64)
            self.charge_state_model["shortest_half_life"] = np.zeros((n_cand,), np.float64)
            row_idx = 0
            for cand in candidates:
                self.charge_state_model["nuclide_hash"][row_idx, 0:len(cand.nuclide_hash)] \
                    = cand.nuclide_hash
                self.charge_state_model["charge_state"][row_idx] \
                    = cand.charge_state
                self.charge_state_model["mass"][row_idx] \
                    = cand.mass
                self.charge_state_model["natural_abundance_product"][row_idx] \
                    = cand.abundance_product
                self.charge_state_model["shortest_half_life"][row_idx] \
                    = cand.shortest_half_life
                row_idx += 1



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Reader for ranging defs extracted from FAU/Erlangen Atom Probe Toolbox Matlab figures FIG.TXT."""

# pylint: disable=too-many-locals

import re
import numpy as np

from ase.data import atomic_numbers
from ifes_apt_tc_data_modeling.nexus.nx_ion import NxField, NxIon
from ifes_apt_tc_data_modeling.utils.definitions import \
    MAX_NUMBER_OF_ATOMS_PER_ION
from ifes_apt_tc_data_modeling.utils.molecular_ions import \
    get_chemical_symbols, isotope_to_hash


class ReadFigTxtFileFormat():
    """Read *.fig.txt file format."""

    def __init__(self, file_path: str):
        if (len(file_path) <= 8) or (file_path.lower().endswith(".fig.txt") is False):
            raise ImportError("WARNING::FIG.TXT file incorrect file_path ending or file type!")
        self.file_path = file_path
        self.fig: dict = {"ranges": {},
                          "ions": {},
                          "molecular_ions": []}
        self.read_fig_txt()

    def read_fig_txt(self):
        """Read FIG.TXT range file content."""
        with open(self.file_path, mode="r", encoding="utf8") as figf:
            txt = figf.read()

        txt = txt.replace("\r\n", "\n")  # windows to unix EOL conversion
        txt = txt.replace(",", ".")  # use decimal dots instead of comma
        txt_stripped = [line for line in txt.split("\n")
                        if line.strip() != "" and line.startswith("#") is False]
        for molecular_ion in txt_stripped:
            tmp = molecular_ion.split(" ")
            mqmin = np.float64(tmp[len(tmp) - 2:-1][0])
            mqmax = np.float64(tmp[len(tmp) - 1:][0])
            ionname = " ".join(tmp[:-2])
            # print(f"{ionname} [{mqmin}, {mqmax}]")
            # ionname = '16O 1H2 + + +  + '

            positive = ionname.count("+")
            negative = ionname.count("-")
            if (0 < positive <= 7) and (negative == 0):
                charge_state = positive
            elif (0 < negative <= 7) and (positive == 0):
                charge_state = -negative
            else:
                charge_state = 0

            tmp = ionname.replace("+", "").replace("-", "").split(" ")
            ivec = []
            for isotope in tmp:
                if isotope != "":
                    prefix = re.findall("^[0-9]+", isotope)
                    mass_number = 0
                    if len(prefix) == 1:
                        if int(prefix[0]) > 0:
                            mass_number = int(prefix[0])
                    suffix = re.findall("[0-9]+$", isotope)
                    multiplier = 1
                    if len(suffix) == 1:
                        multiplier = int(suffix[0])
                    symbol = isotope.replace(
                        f"{mass_number}", "").replace(f"{multiplier}", "").replace(" ", "")
                    if symbol in get_chemical_symbols():
                        proton_number = atomic_numbers[symbol]
                        neutron_number = 0
                        if mass_number != 0:
                            neutron_number = mass_number - proton_number
                        ivec.extend([isotope_to_hash(proton_number, neutron_number)] * multiplier)
            ivec = np.sort(np.asarray(ivec, np.uint16))[::-1]
            ivector = np.zeros((MAX_NUMBER_OF_ATOMS_PER_ION,), np.uint16)
            ivector[0:len(ivec)] = ivec

            m_ion = NxIon(nuclide_hash=ivector, charge_state=charge_state)
            m_ion.add_range(mqmin, mqmax)
            m_ion.comment = NxField(ionname, "")
            m_ion.apply_combinatorics()
            # m_ion.report()

            self.fig["molecular_ions"].append(m_ion)

        print(f"{self.file_path} parsed successfully")
